#include <stdio.h>
#include <string.h>

#include <math.h>

#include "matrixmath.h"

static void MatrixMultiRotate3(double *a, double *b, double *r);

void MatrixRotateHeading3(double *a, double theta, double *r);
void MatrixRotatePitch3(double *a, double theta, double *r);
void MatrixRotateBank3(double *a, double theta, double *r);


/*
 *	Multiplies matrix a (3 by 1) with rotational matrix b (3 by 3)
 *	and stores results into matrix r (3 by 1).
 */
static void MatrixMultiRotate3(double *a, double *b, double *r)
{
	int i, j;
	int m = 3, n = 3;

	for(j = 0; j < n; j++)
	{
	    r[j] = 0.0;
	    for(i = 0; i < m; i++)
		r[j] += a[i] * b[(i * m) + j];
	}

	return;
}

/*
 *	Rotates position matrix a (3 * 1) about z axis clockwise.
 *
 *	Theta is in radians.
 */
void MatrixRotateHeading3(double *a, double theta, double *r)
{
	double b[3 * 3];

        b[0] = cos(theta);
        b[1] = -sin(theta);
        b[2] = 0.0;

        b[3] = -b[1];	/* -sin(theta) */
        b[4] = b[0];	/* cos(theta) */
        b[5] = 0.0;

        b[6] = 0.0;
        b[7] = 0.0;
        b[8] = 1.0;

	MatrixMultiRotate3(a, b, r);	

	return;
}

/*
 *	Rotates position matrix a (3 * 1) about x axis clockwise.
 *
 *	Theta is in radians.
 */
void MatrixRotatePitch3(double *a, double theta, double *r)
{
        double b[3 * 3];

        b[0] = 1.0;
        b[1] = 0.0;
        b[2] = 0.0;  

        b[3] = 0.0;
        b[4] = cos(theta);
        b[5] = -sin(theta);

        b[6] = 0.0;
        b[7] = -b[5];	/* sin(theta) */
        b[8] = b[4];	/* cos(theta) */

        MatrixMultiRotate3(a, b, r);

        return;
}       

/*
 *	Rotates position matrix a (3 * 1) about -y axis (netagive y axis)
 *	clockwise.
 *
 *	Theta is in radians.
 */
void MatrixRotateBank3(double *a, double theta, double *r)
{
        double b[3 * 3];

        b[0] = cos(-theta);
        b[1] = 0.0;
        b[2] = sin(-theta);
        
        b[3] = 0.0;
        b[4] = 1.0;
        b[5] = 0.0;

        b[6] = -b[2];	/* -sin(-theta) */
        b[7] = 0.0;
        b[8] = b[0];	/* cos(-theta) */

        MatrixMultiRotate3(a, b, r);

        return;
}
