#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <math.h>

#include "../include/os.h"
#ifdef __MSW__
# include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>

#include "../include/string.h"

#include "sfm.h"

#include "gw.h"
#include "stategl.h"
#include "textinput.h"

#include "sarreality.h"
#include "obj.h"
#include "messages.h"
#include "gctl.h"
#include "sarsound.h"

#include "sar.h"
#include "config.h"


void SARDrawGetDirFromPos(
	sar_position_struct *pos1, sar_position_struct *pos2,
	sar_direction_struct *dir_rtn
);
void SARDrawSetColor(sar_color_struct *c);
void SARVertexBox(double x_len, double y_len, double z_len);
void SARVertexBoxBase(double x_len, double y_len, double z_len);

void SARSetCamera(
	sar_scene_struct *scene,
	sar_object_struct ***ptr, int *total
);
void SARDrawSetSpotLight(
	gw_display_struct *display, sar_scene_struct *scene,
	sar_object_struct *obj_ptr
);
void SARDrawHelp(sar_core_struct *core_ptr);

void SARDrawRotor(
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_obj_rotor_struct *rotor_ptr
);
void SARDrawLandingGear(
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_obj_landing_gear_struct *lgear_ptr
);
void SARDrawDoor(
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_obj_door_struct *door_ptr
);
void SARDrawHoist(
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_obj_hoist_struct *hoist_ptr,
	double rope_len
);
void SARDrawSmoke(
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_object_smoke_struct *obj_smoke_ptr
);
void SARDrawExplosion(
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
        sar_object_explosion_struct *obj_explosion_ptr
);
void SARDrawHelipad(
	gw_display_struct *display, sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
        sar_object_helipad_struct *obj_helipad_ptr,
	double distance
);
void SARDrawRunway(
	gw_display_struct *display, sar_scene_struct *scene,
        sar_object_struct *obj_ptr,
        sar_object_runway_struct *obj_runway_ptr,
        double distance
);
void SARDrawHuman(
	gw_display_struct *display, sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_object_human_struct *obj_human_ptr
);

void SARDrawHUDAttitude(
        gw_display_struct *display, sar_scene_struct *scene,
        sar_direction_struct *dir, sar_position_struct *vel,
        int offset_x, int offset_y,
        int width, int height,
        double fovx, double fovy
);
void SARDrawHUDHeading(
	gw_display_struct *display, sar_scene_struct *scene,
	sar_direction_struct *dir, sar_obj_intercept_struct *intercept,
        int offset_x, int offset_y,
        int width, int height
);
void SARDrawHUD(
        gw_display_struct *display, sar_scene_struct *scene,
        sar_object_struct ***ptr, int *total,
        double aspect
);
void SARDrawOutsideAttitude(
        gw_display_struct *display, sar_scene_struct *scene,
        sar_object_struct *obj_ptr,
	int width, int height
);

void SARDrawMessages(
	gw_display_struct *display, sar_scene_struct *scene
);

void SARDrawLights(
	gw_display_struct *display, sar_scene_struct *scene,
	sar_light_struct **ptr, int total
);

void SARDrawObjectCockpit(
	sar_scene_struct *scene,  sar_object_struct *obj_ptr
);

void SARDrawObjectShadow(
	gw_display_struct *display, sar_scene_struct *scene,
	sar_object_struct *obj_ptr
);

void SARDrawSceneFoundations(
	gw_display_struct *display, sar_scene_struct *scene,
	double aspect
);

void SARDrawCloudLayer(
	sar_scene_struct *scene, sar_cloud_layer_struct *cloud_layer_ptr,
        Boolean flip
);
void SARDrawSceneHorizon(
	gw_display_struct *display, sar_scene_struct *scene
);

void SARDraw(sar_core_struct *core_ptr);
void SARDrawMap(sar_core_struct *core_ptr, Boolean draw_for_gcc);
        

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))


/* Get camera orientation back from GL? */
#define GET_CAM_DIR	False


/* Turns on nessesary GL dependant states just before drawing point lights. */
#define STATE_FOR_POINT_LIGHTS_BEGIN	\
{ \
 StateGLDisable(&display->state_gl, GL_LIGHTING); \
 StateGLDisable(&display->state_gl, GL_LIGHT0); \
}
/* Turns off nessesary GL states just before drawing point lights. */
#define STATE_FOR_POINT_LIGHTS_END	\
{ \
 StateGLEnable(&display->state_gl, GL_LIGHTING); \
 StateGLEnable(&display->state_gl, GL_LIGHT0); \
}


/* Turns on GL_DEPTH_TEST and sets the DepthFunc. */
#define SAR_DRAW_DEPTH_TEST_ON	\
{ \
 StateGLEnable(&display->state_gl, GL_DEPTH_TEST); \
 StateGLDepthFunc(&display->state_gl, GL_LEQUAL); \
}
/* Turns off GL_DEPTH_TEST and sets the DepthFunc. */
#define SAR_DRAW_DEPTH_TEST_OFF	\
{ \
 StateGLDisable(&display->state_gl, GL_DEPTH_TEST); \
 StateGLDepthFunc(&display->state_gl, GL_ALWAYS); \
}

/* Turns on GL_TEXTURE_1D and sets the TexEnv. */
#define SAR_DRAW_TEXTURE_1D_ON  \
{ \
 StateGLEnable(&display->state_gl, GL_TEXTURE_1D); \
 StateGLTexEnvI( \
  &display->state_gl, \
  GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE \
 ); \
}
/* Turns off GL_TEXTURE_1D. */
#define SAR_DRAW_TEXTURE_1D_OFF \
{ \
 StateGLDisable(&display->state_gl, GL_TEXTURE_1D); \
}

/* Turns on GL_TEXTURE_2D and sets the TexEnv. */
#define SAR_DRAW_TEXTURE_2D_ON  \
{ \
 StateGLEnable(&display->state_gl, GL_TEXTURE_2D); \
 StateGLTexEnvI( \
  &display->state_gl, \
  GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE \
 ); \
}
/* Turns off GL_TEXTURE_2D. */
#define SAR_DRAW_TEXTURE_2D_OFF \
{ \
 StateGLDisable(&display->state_gl, GL_TEXTURE_2D); \
}


/* Force disable of certain states which might have been enabled during
 * a SARVisualModelCallList().
 */
#define POST_CALLLIST_RESET_STATES	\
{ \
 StateGLDisableF(&display->state_gl, GL_BLEND, GL_TRUE); \
}


static double visibility_max;           /* In meters. */
static int player_wheel_brakes, player_air_brakes;
static Boolean camera_in_cockpit;
static int camera_ref;
static sar_position_struct camera_pos; 
static sar_direction_struct camera_dir;
static GLfloat sky_color[3], atmosphere_color[3];
static sar_cloud_layer_struct   *lowest_cloud_layer_ptr,    
                                *highest_cloud_layer_ptr;


/*
 *	Calculates the heading and pitch of position pos1 to
 *	position pos2, putting the results in dir_rtn.
 */
void SARDrawGetDirFromPos(
        sar_position_struct *pos1, sar_position_struct *pos2,
        sar_direction_struct *dir_rtn
)
{
	double r;
	sar_position_struct delta;

	if((pos1 == NULL) || (pos2 == NULL) || (dir_rtn == NULL))
	    return;

	delta.x = pos2->x - pos1->x;
	delta.y = pos2->y - pos1->y;
	delta.z = pos2->z - pos1->z;
	r = hypot(delta.x, delta.y);

	dir_rtn->heading = SFMSanitizeRadians(
	    (0.5 * PI) - atan2(delta.y, delta.x)
	);
        dir_rtn->pitch = SFMSanitizeRadians(
            -atan2(delta.z, r)
        );
	dir_rtn->bank = 0.0;

	return;
}


/*
 *	Set gl color from color structure.
 */
void SARDrawSetColor(sar_color_struct *c)
{
	if(c == NULL)
	    return;

	glColor4f(
	    (GLfloat)c->r,
	    (GLfloat)c->g,
	    (GLfloat)c->b,
	    (GLfloat)c->a
	);

	return;
}

/*
 *	Sets vertexes to create a 3D box at the center.
 */
void SARVertexBox(
	double x_len, double y_len, double z_len
)
{
	double x_min, x_max;
	double y_min, y_max;
	double z_min, z_max;

	x_max = x_len / 2;
	x_min = -x_max;

        y_max = y_len / 2;
        y_min = -y_max;

        z_max = z_len / 2;
        z_min = -z_max;

	/* Front and back. */
	glNormal3f(0, 0, -1);
	glVertex3f((GLfloat)x_min, (GLfloat)z_min, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_min, (GLfloat)z_max, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_max, (GLfloat)z_max, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_max, (GLfloat)z_min, (GLfloat)-y_max);

        glNormal3f(0, 0, 1);
        glVertex3f((GLfloat)x_min, (GLfloat)z_max, (GLfloat)-y_min);
        glVertex3f((GLfloat)x_min, (GLfloat)z_min, (GLfloat)-y_min);
        glVertex3f((GLfloat)x_max, (GLfloat)z_min, (GLfloat)-y_min);
        glVertex3f((GLfloat)x_max, (GLfloat)z_max, (GLfloat)-y_min);

	/* Left and right. */
        glNormal3f(-1, 0, 0);
        glVertex3f((GLfloat)x_min, (GLfloat)z_max, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_min, (GLfloat)z_min, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_min, (GLfloat)z_min, (GLfloat)-y_min);
        glVertex3f((GLfloat)x_min, (GLfloat)z_max, (GLfloat)-y_min);

        glNormal3f(1, 0, 0);
        glVertex3f((GLfloat)x_max, (GLfloat)z_min, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_max, (GLfloat)z_max, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_max, (GLfloat)z_max, (GLfloat)-y_min);
        glVertex3f((GLfloat)x_max, (GLfloat)z_min, (GLfloat)-y_min);

	/* Top and bottom. */
        glNormal3f(0, 1, 0);
        glVertex3f((GLfloat)x_min, (GLfloat)z_max, (GLfloat)-y_min);
        glVertex3f((GLfloat)x_max, (GLfloat)z_max, (GLfloat)-y_min);
        glVertex3f((GLfloat)x_max, (GLfloat)z_max, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_min, (GLfloat)z_max, (GLfloat)-y_max);

        glNormal3f(0, -1, 0);
        glVertex3f((GLfloat)x_min, (GLfloat)z_min, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_max, (GLfloat)z_min, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_max, (GLfloat)z_min, (GLfloat)-y_min);
        glVertex3f((GLfloat)x_min, (GLfloat)z_min, (GLfloat)-y_min);

	return;
}

/*
 *      Sets vertexes to create a 3D box in the positive octant
 *	with respect to the x y plane.  The x and y walls are
 *	centered. Bottom of box is not closed.
 */
void SARVertexBoxBase(
        double x_len, double y_len, double z_len
)
{
        double x_min, x_max;
        double y_min, y_max;
	double z_min, z_max;

        x_max = x_len / 2;
        x_min = -x_max;

        y_max = y_len / 2;
        y_min = -y_max;

	z_min = 0;
	z_max = z_len;

        /* Front and back. */
        glNormal3f(0, 0, -1);
        glVertex3f((GLfloat)x_min, (GLfloat)z_min, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_min, (GLfloat)z_max, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_max, (GLfloat)z_max, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_max, (GLfloat)z_min, (GLfloat)-y_max);
            
        glNormal3f(0, 0, 1);
        glVertex3f((GLfloat)x_max, (GLfloat)z_min, (GLfloat)-y_min);
        glVertex3f((GLfloat)x_max, (GLfloat)z_max, (GLfloat)-y_min);
        glVertex3f((GLfloat)x_min, (GLfloat)z_max, (GLfloat)-y_min);
        glVertex3f((GLfloat)x_min, (GLfloat)z_min, (GLfloat)-y_min);

        /* Left and right. */
        glNormal3f(-1, 0, 0);
        glVertex3f((GLfloat)x_min, (GLfloat)z_min, (GLfloat)-y_min);
        glVertex3f((GLfloat)x_min, (GLfloat)z_max, (GLfloat)-y_min);
        glVertex3f((GLfloat)x_min, (GLfloat)z_max, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_min, (GLfloat)z_min, (GLfloat)-y_max);
        
        glNormal3f(1, 0, 0);
        glVertex3f((GLfloat)x_max, (GLfloat)z_min, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_max, (GLfloat)z_max, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_max, (GLfloat)z_max, (GLfloat)-y_min);
        glVertex3f((GLfloat)x_max, (GLfloat)z_min, (GLfloat)-y_min);

        /* Top only. */
        glNormal3f(0, 1, 0);
        glVertex3f((GLfloat)x_max, (GLfloat)z_max, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_min, (GLfloat)z_max, (GLfloat)-y_max);
        glVertex3f((GLfloat)x_min, (GLfloat)z_max, (GLfloat)-y_min);
        glVertex3f((GLfloat)x_max, (GLfloat)z_max, (GLfloat)-y_min);

	return;
}

/*
 *	Translates camera to proper position, also updates globals
 *	camera_pos, camera_dir, and camera_in_cockpit.
 *
 *	Note, if GET_CAM_DIR is False then camera_dir will be all
 *	zero'ed.
 */
void SARSetCamera(
	sar_scene_struct *scene,
	sar_object_struct ***ptr,
	int *total
)
{
	int n;
	double x, y, z;
	sar_object_struct *obj_ptr;
	sar_direction_struct *dir, *dir2;
	sar_position_struct *pos, *pos2;
	sar_object_aircraft_struct *obj_aircraft_ptr = NULL;
	GLdouble val[16];


	/* Update camera ref code. */
	camera_ref = scene->camera_ref;

	switch(camera_ref)
	{
          case SAR_CAMERA_REF_HOIST:
            camera_in_cockpit = True;
            obj_ptr = scene->player_obj_ptr;
            if(obj_ptr != NULL)
            {
                double sin_heading, cos_heading;
                double cos_pitch;
                double x_offset, y_offset, z_offset;
		sar_obj_hoist_struct *hoist_ptr;


		/* Position camera directly behind hoist, direction
		 * relative to object heading.
		 */
                if(obj_ptr->type == SAR_OBJ_TYPE_AIRCRAFT)
                    obj_aircraft_ptr = obj_ptr->data;
                if(obj_aircraft_ptr == NULL)
                    break;

		hoist_ptr = &obj_aircraft_ptr->hoist;

		if(hoist_ptr->rope_cur > 0)
		{
                    pos = &hoist_ptr->pos;
                    dir = &obj_ptr->dir;
                    dir2 = &scene->camera_hoist_dir;
		}
		else
		{
		    pos = &obj_ptr->pos;
                    dir = &obj_ptr->dir;
                    dir2 = &scene->camera_hoist_dir;
		}

                /* Pre calculate object heading trig values. */
                sin_heading = sin(dir->heading);
                cos_heading = cos(dir->heading);
                cos_pitch = cos(dir2->pitch);   /* Camera spot pitch. */

                /* Spot camera offset. */
                x_offset = sin(dir2->heading) * cos_pitch *
                    scene->camera_hoist_dist; 
                y_offset = cos(dir2->heading) * cos_pitch *
                    scene->camera_hoist_dist;
                z_offset = -sin(dir2->pitch) * scene->camera_hoist_dist;

                x = pos->x + (cos_heading * x_offset) +
                    (sin_heading * y_offset);
                y = pos->y - (sin_heading * x_offset) +
                    (cos_heading * y_offset);
                z = pos->z + z_offset;
                if(z < 0)
                    z = 0;
                gluLookAt(
                    x, z, -y,
                    pos->x, pos->z, -pos->y,
                    0, 1, 0
                );

                /* Record camera position. */
                camera_pos.x = x;
                camera_pos.y = y;   
                camera_pos.z = z;

                memcpy(&scene->camera_pos, &camera_pos, sizeof(sar_position_struct));
	    }
            break;  

	  case SAR_CAMERA_REF_MAP:
            camera_in_cockpit = False;

            obj_ptr = scene->player_obj_ptr;
            if(obj_ptr != NULL)
            {
                pos = &scene->camera_pos;

                /* Set up initial normal before we translate. */
/*		glNormal3f(0, 0, 1); */

                glRotatef(90, 1.0, 0.0, 0.0);
                glTranslatef((GLfloat)-pos->x, (GLfloat)-pos->z, (GLfloat)pos->y);

                /* Record camera position. */
                camera_pos.x = pos->x;
                camera_pos.y = pos->y;
                camera_pos.z = pos->z;

                /* Leave scene camera position where it is at. */
	    }
	    break;

	  case SAR_CAMERA_REF_TOWER:
	    camera_in_cockpit = False;

            n = scene->camera_target;
            if(SARObjIsAllocated(*ptr, *total, n))
            {
		double x, y, z;


                obj_ptr = (*ptr)[n];

		x = scene->camera_pos.x;
		y = scene->camera_pos.y;
		z = scene->camera_pos.z;
		if(z < 0)
		    z = 0;
                pos = &obj_ptr->pos;

                gluLookAt(
                    x, z, -y,
                    pos->x, pos->z, -pos->y,
                    0, 1, 0
                );

                /* Record camera position. */
                camera_pos.x = scene->camera_pos.x;
                camera_pos.y = scene->camera_pos.y;
                camera_pos.z = scene->camera_pos.z;

		/* Leave scene camera position where it is at. */
	    }
	    break;

	  case SAR_CAMERA_REF_SPOT:
	    camera_in_cockpit = False;

	    n = scene->camera_target;
	    if(SARObjIsAllocated(*ptr, *total, n))
	    {
		double sin_heading, cos_heading;
		double cos_pitch;
		double x_offset, y_offset, z_offset;


	        obj_ptr = (*ptr)[n];

                pos = &obj_ptr->pos;
                dir = &obj_ptr->dir;
		dir2 = &scene->camera_spot_dir;

		/* Pre calculate object heading trig values. */
		sin_heading = sin(dir->heading);
		cos_heading = cos(dir->heading);
		cos_pitch = cos(dir2->pitch);	/* Camera spot pitch. */

		/* Spot camera offset. */
		x_offset = sin(dir2->heading) * cos_pitch *
                    scene->camera_spot_dist;
		y_offset = cos(dir2->heading) * cos_pitch *
		    scene->camera_spot_dist;
		z_offset = -sin(dir2->pitch) * scene->camera_spot_dist;

		x = pos->x + (cos_heading * x_offset) +
		    (sin_heading * y_offset);
		y = pos->y - (sin_heading * x_offset) +
		    (cos_heading * y_offset);
		z = pos->z + z_offset;
		/* Keep z above ground. */
		if(z < 0.5)
		    z = 0.5;
		gluLookAt(
		    x, z, -y,
		    pos->x, pos->z, -pos->y,
		    0, 1, 0
		);

		/* Record camera position. */
		camera_pos.x = x;
		camera_pos.y = y;
		camera_pos.z = z;

		memcpy(&scene->camera_pos, &camera_pos, sizeof(sar_position_struct));
	    }
	    break;

	  default:	/* Default to cockpit. */
	    camera_in_cockpit = True;

	    obj_ptr = scene->player_obj_ptr;
            if(obj_ptr != NULL)
            {
		if(obj_ptr->type == SAR_OBJ_TYPE_AIRCRAFT)
		    obj_aircraft_ptr = obj_ptr->data;
		if(obj_aircraft_ptr == NULL)
		    break;

                pos = &obj_ptr->pos;
                dir = &obj_ptr->dir;

		pos2 = &obj_aircraft_ptr->cockpit_offset_pos;
                dir2 = &scene->camera_cockpit_dir;

		/* Set up initial normal before we translate. */
		glNormal3f(0, 0, 1);

                glRotatef((GLfloat)SFMRadiansToDegrees(dir2->pitch), 1.0, 0.0, 0.0);
                glRotatef((GLfloat)SFMRadiansToDegrees(dir2->heading), 0.0, 1.0, 0.0);
                glTranslatef((GLfloat)-pos2->x, (GLfloat)-pos2->z, (GLfloat)pos2->y);

                glRotatef((GLfloat)SFMRadiansToDegrees(dir->bank), 0.0, 0.0, 1.0);
                glRotatef((GLfloat)SFMRadiansToDegrees(dir->pitch), 1.0, 0.0, 0.0);
                glRotatef((GLfloat)SFMRadiansToDegrees(dir->heading), 0.0, 1.0, 0.0);
                glTranslatef((GLfloat)-pos->x, (GLfloat)-pos->z, (GLfloat)pos->y);

                /* Record camera position. */
                camera_pos.x = pos->x;
                camera_pos.y = pos->y;
                camera_pos.z = pos->z;

                memcpy(&scene->camera_pos, &camera_pos, sizeof(sar_position_struct));
	    }
	    break;
        }

	/* Update camera direction? */
	if(GET_CAM_DIR)
	{
            /* Get matrix containing results of our camera rotations
	     * and translations.
	     */
            glGetDoublev(GL_MODELVIEW_MATRIX, val);

	    /* Record camera direction. */
	    camera_dir.heading = SFMSanitizeRadians(
	        (0.5 * PI) - atan2(val[10], val[8])
            );
	    camera_dir.pitch = SFMSanitizeRadians(
	        atan2(val[6], val[5])
	    );
	    camera_dir.bank = SFMSanitizeRadians(asin(-val[4]));
	}
	else
	{
	    memset(&camera_dir, 0x00, sizeof(sar_direction_struct));
	}

	return;
}

/*
 *	Sets up the search light (GL_LIGHT1) for the player object.
 *
 *	Inputs assumed valid.
 *
 *	Translation matrix should be translated to light position.
 */
void SARDrawSetSpotLight(
	gw_display_struct *display,
	sar_scene_struct *scene, sar_object_struct *obj_ptr
)
{
	int i;
	Boolean state = False;
	int light_num = GL_LIGHT1;
	sar_position_struct pos;
	double scene_g, g = 1.0;
	sar_object_aircraft_struct *obj_aircraft_ptr = NULL;
	sar_light_struct *light_ptr;


	/* Get scene gamma as just the red compient of the scene light. */
	scene_g = scene->light_color.r;

	/* Handle by object type. */
	switch(obj_ptr->type)
	{
	  case SAR_OBJ_TYPE_AIRCRAFT:
	    obj_aircraft_ptr = obj_ptr->data;
	    if(obj_aircraft_ptr == NULL)
		break;

	    /* Scene gamma less than 0.75? */
	    if(scene_g < 0.75)
	    {
		/* Object has a spot light? */
		for(i = 0; i < obj_ptr->total_lights; i++)
		{
		    light_ptr = obj_ptr->light[i];
		    if(light_ptr == NULL)
			continue;

		    if(light_ptr->flags & SAR_LIGHT_FLAG_ATTENUATE)
		    {
			if(light_ptr->flags & SAR_LIGHT_FLAG_ON)
			{
			    /* Has light, set state to True and get
			     * light position relative to center of
			     * object.
			     */
			    state = True;
			    memcpy(
				&pos, &light_ptr->pos,
				sizeof(sar_position_struct)
			    );
			}
			break;
		    }
		}

		/* Calculate gamma for spot light, as scene gamma
		 * gets darker, the spot light gets brighter.
		 */
		g = CLIP(1.0 - scene_g, 0.0, 1.0);
	    }
	    break;
	}

	/* If stat is True then that implies a spot light is on. */
	if(state)
	{
	    GLfloat v[4];

            StateGLEnable(&display->state_gl, light_num);

	    v[0] = (GLfloat)pos.x;
	    v[1] = (GLfloat)pos.z;
	    v[2] = (GLfloat)-pos.y;
	    v[3] = 1.0;
	    glLightfv(light_num, GL_POSITION, &(v[0]));

            v[0] = 0.0;
	    v[1] = 0.0;
	    v[2] = 0.0;
	    v[3] = 1.0;
	    glLightfv(light_num, GL_AMBIENT, &(v[0]));

            v[0] = (GLfloat)g;
            v[1] = (GLfloat)g;
            v[2] = (GLfloat)g;
            v[3] = 1.0;
            glLightfv(light_num, GL_DIFFUSE, &(v[0]));

	    glLightf(light_num, (GLuint)GL_CONSTANT_ATTENUATION, (GLfloat)1.0);
            glLightf(light_num, (GLuint)GL_LINEAR_ATTENUATION, (GLfloat)0.01);
	    glLightf(light_num, (GLuint)GL_QUADRATIC_ATTENUATION, (GLfloat)0.0);
            v[0] = (GLfloat)0.0;	/* x. */
            v[1] = (GLfloat)-0.6;	/* z. */
            v[2] = (GLfloat)-0.3;	/* y. */
            glLightfv(light_num, GL_SPOT_DIRECTION, &(v[0]));
            glLightf(light_num, GL_SPOT_CUTOFF, 60.0);
	    glLightf(light_num, GL_SPOT_EXPONENT, 0.0);
	}
	else
	{
	    StateGLDisable(&display->state_gl, light_num);
	}

	return;
}

/*
 *	Draws help message.
 */
void SARDrawHelp(sar_core_struct *core_ptr)
{
	gw_display_struct *display;
	int i, xc, yc, ymax, fw, fh, width, height;
        GWFont *font;
        sar_color_struct *color;
	char *mesg[] = SAR_KEYS_HELP_MESSAGES;
	int mesg_lines, mesg_line_max = SAR_KEYS_HELP_MESSAGE_LINE_MAX;


	if(core_ptr == NULL)
	    return;

	display = core_ptr->display;
	if(display == NULL)
	    return;

	width = display->width;
	height = display->height;

        color = &option.message_color;
        font = option.message_font;
        if(font == NULL)
            return;

        GWGetFontSize(font, NULL, NULL, &fw, &fh);
	if((fw <= 0) ||
           (fh <= 0)
	)
	    return;

	/* Calculate line arragements. */
        mesg_lines = (sizeof(mesg) / sizeof(char *));

        /* Draw shaded background. */
        StateGLEnable(&display->state_gl, GL_BLEND);
        StateGLBlendFunc(
	    &display->state_gl, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA
	);
        glBegin(GL_QUADS);
        {
            glColor4f(0.0, 0.0, 0.0, 0.5);
            glVertex2f(0, 0);
            glVertex2f((GLfloat)width, (GLfloat)0);
            glVertex2f((GLfloat)width, (GLfloat)height);
            glVertex2f((GLfloat)0, (GLfloat)height);
        }
        glEnd();
        StateGLDisable(&display->state_gl, GL_BLEND);


        /* Set text color and font. */
        glColor4f(
            (GLfloat)color->r,
	    (GLfloat)color->g,
	    (GLfloat)color->b,
	    (GLfloat)color->a
        );
        GWSetFont(display, font);

        /* Draw text. */
	xc = 10; yc = 5; ymax = height - fh - 5;
        GWDrawString(display, xc, yc, "Keyboard Functions:");
	yc = fh + (2 * 5);
	i = 0;
	while((yc < ymax) && (i < mesg_lines))
	{
            GWDrawString(display, xc, yc, mesg[i]);
	    yc += fh; i++;
	}

	xc += (mesg_line_max * fw);
        yc = fh + (2 * 5);
        while((yc < ymax) && (i < mesg_lines))
        {
            GWDrawString(display, xc, yc, mesg[i]);
            yc += fh; i++;
        }

	return;
}
 

/*
 *      Draws the rotor/propellar relative to the given object.
 *
 *      Inputs assumed valid.
 */
void SARDrawRotor(
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_obj_rotor_struct *rotor_ptr
)
{
        sar_position_struct *pos;
        sar_direction_struct *dir;	/* In radians. */
        double anim_coeff;


	if(rotor_ptr->visual_model == NULL)
	    return;

	/* Does not rotate? */
	if(rotor_ptr->flags & SAR_ROTOR_FLAG_NO_ROTATE)
	    anim_coeff = 0;
	else
	    anim_coeff = (double)rotor_ptr->anim_pos /
		(double)((sar_grad_anim_t)-1);

        /* Translation assumed at object's center position. */
        glPushMatrix();
        {
            /* Get offset from center of object. */
            pos = &rotor_ptr->pos;
            glTranslatef((GLfloat)pos->x, (GLfloat)pos->z, (GLfloat)-pos->y);

            /* Get direction. */
            dir = &rotor_ptr->dir;
            glRotatef(
                (GLfloat)-SFMRadiansToDegrees(dir->heading),
                0.0, 1.0, 0.0
            );
	    if(rotor_ptr->flags & SAR_ROTOR_FLAG_PITCH_STATE)
		glRotatef(
                    (GLfloat)-SFMRadiansToDegrees(dir->pitch + (0.5 * PI)),
                    1.0, 0.0, 0.0
                );
	    else
                glRotatef(
                    (GLfloat)-SFMRadiansToDegrees(dir->pitch),
                    1.0, 0.0, 0.0
                );
            glRotatef(
                (GLfloat)-SFMRadiansToDegrees(dir->bank),
                0.0, 0.0, 1.0
            );

	    glPushMatrix();
	    {
		/* Rotate counter clockwise along Z axis. */
                glRotatef(
                    (GLfloat)-SFMRadiansToDegrees(-anim_coeff * (2 * PI)),
                    0.0, 1.0, 0.0
                );
		SARVisualModelCallList(rotor_ptr->visual_model);
	    }
	    glPopMatrix();
        }
        glPopMatrix();

	return;
}

/*
 *	Draws the landing gear relative to the given object.
 *
 *	Inputs assumed valid.
 */
void SARDrawLandingGear(
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_obj_landing_gear_struct *lgear_ptr
)
{
	sar_position_struct *pos;
	sar_direction_struct *dir;	/* In radians. */
	double anim_coeff;


	if(lgear_ptr->visual_model == NULL)
	    return;

	anim_coeff = (double)lgear_ptr->anim_pos /
	    (double)((sar_grad_anim_t)-1);

	/* A coeff of 0.0 implies all retracted. */
	if(anim_coeff <= 0.0)
	    return;

	/* Translation assumed at object's center position. */
	glPushMatrix();
	{
	    /* Get offset from center of object. */
            pos = &lgear_ptr->pos;
            glTranslatef((GLfloat)pos->x, (GLfloat)pos->z, (GLfloat)-pos->y);

	    /* Calculate direction. */
            dir = &lgear_ptr->extended_dir;

	    glRotatef(
		(GLfloat)-SFMRadiansToDegrees(dir->heading * anim_coeff),
                0.0, 1.0, 0.0
	    );
            glRotatef(
		(GLfloat)-SFMRadiansToDegrees(dir->pitch * anim_coeff),
		1.0, 0.0, 0.0
	    );
            glRotatef(
		(GLfloat)-SFMRadiansToDegrees(dir->bank * anim_coeff),
		0.0, 0.0, 1.0
	    );
	    SARVisualModelCallList(lgear_ptr->visual_model);
        }
	glPopMatrix();

	return;
}

/*
 *	Draws door on the object.
 */
void SARDrawDoor(
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_obj_door_struct *door_ptr
)
{
        sar_position_struct *pos_closed, *pos_opened, delta;
        sar_direction_struct *dir;
        double anim_coeff;


	if(door_ptr->visual_model == NULL)
	    return;

        anim_coeff = (double)door_ptr->anim_pos /
            (double)((sar_grad_anim_t)-1);

	pos_closed = &door_ptr->pos_closed;
	pos_opened = &door_ptr->pos_opened;
	dir = &door_ptr->dir_opened;

	delta.x = pos_opened->x - pos_closed->x;
	delta.y = pos_opened->y - pos_closed->y;
	delta.z = pos_opened->z - pos_closed->z;

	delta.x = pos_closed->x + (delta.x * anim_coeff);
	delta.y = pos_closed->y + (delta.y * anim_coeff);
	delta.z = pos_closed->z + (delta.z * anim_coeff);

        /* Translation assumed at object's center position. */
        glPushMatrix();
        {
            /* Translate from center of object. */
            glTranslatef((GLfloat)delta.x, (GLfloat)delta.z, (GLfloat)-delta.y);

            /* Rotate to opened position. */
            glRotatef(
                (GLfloat)-SFMRadiansToDegrees(dir->heading * anim_coeff),
                0.0, 1.0, 0.0
            );
            glRotatef(
                (GLfloat)-SFMRadiansToDegrees(dir->pitch * anim_coeff),
                1.0, 0.0, 0.0
            );
            glRotatef(
                (GLfloat)-SFMRadiansToDegrees(dir->bank * anim_coeff),
                0.0, 0.0, 1.0
            );
	    SARVisualModelCallList(door_ptr->visual_model);
        }
        glPopMatrix();

	return;
}

/*
 *	Draws hoist and rescue basket.
 */
void SARDrawHoist(
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_obj_hoist_struct *hoist_ptr,
	double rope_len
)
{
	int tex_num;
	double l, w, h, hl, hw, hh;
	sar_position_struct *pos;
	v3d_texture_ref_struct *side_tex, *end_tex, *bottom_tex;


	pos = &hoist_ptr->pos;
	l = hoist_ptr->contact_radius * 2;
	w = l / 2;
	h = hoist_ptr->contact_z_max;

	hl = l / 2;
	hw = w / 2;
	hh = h / 2;


	/* Get pointers to textures. */
	tex_num = hoist_ptr->side_tex_num;
	if(SARIsTextureAllocated(scene, tex_num))
	    side_tex = scene->texture_ref[tex_num];
	else
	    side_tex = NULL;

	tex_num = hoist_ptr->end_tex_num;
	if(SARIsTextureAllocated(scene, tex_num))
	    end_tex = scene->texture_ref[tex_num];
	else
	    end_tex = NULL;

	tex_num = hoist_ptr->bottom_tex_num;
	if(SARIsTextureAllocated(scene, tex_num))
	    bottom_tex = scene->texture_ref[tex_num];
	else
	    bottom_tex = NULL;


	/* Begin drawing rescue basket. */
	glShadeModel(GL_FLAT);

	/* Draw bottom base. */
	glColor4f(1.0, 1.0, 1.0, 1.0);
	V3DTextureSelect(bottom_tex);
	glBegin(GL_QUADS);
	{
            glNormal3f(0, 1, 0);  
            glTexCoord2f(0, 1 - 0);
            glVertex3f((GLfloat)-hl, 0,  (GLfloat)hw);
            glTexCoord2f(1, 1 - 0);
            glVertex3f((GLfloat)hl, 0,  (GLfloat)hw);
            glTexCoord2f(1, 1 - 1);
            glVertex3f((GLfloat)hl, 0, (GLfloat)-hw);
            glTexCoord2f(0, 1 - 1);
            glVertex3f((GLfloat)-hl, 0, (GLfloat)-hw);

            glNormal3f(0, -1, 0);
	    glTexCoord2f(0, 1 - 1);
            glVertex3f((GLfloat)-hl, 0, (GLfloat)-hw);
	    glTexCoord2f(1, 1 - 1);
            glVertex3f((GLfloat)hl, 0, (GLfloat)-hw);
	    glTexCoord2f(1, 1 - 0);
            glVertex3f((GLfloat)hl, 0,  (GLfloat)hw);
	    glTexCoord2f(0, 1 - 0);
            glVertex3f((GLfloat)-hl, 0,  (GLfloat)hw);
	}
	glEnd();

	/* Sides. */
        V3DTextureSelect(side_tex);
	glBegin(GL_QUADS);
	{
	    /* Front side. */
            glNormal3f(0, 0, -1);   
            glTexCoord2f(1, 1 - 1);
            glVertex3f((GLfloat)-hl, (GLfloat)hh, (GLfloat)-hw);
            glTexCoord2f(0, 1 - 1);
            glVertex3f( (GLfloat)hl, (GLfloat)hh, (GLfloat)-hw);
            glTexCoord2f(0, 1 - 0);
            glVertex3f( (GLfloat)hl, 0,           (GLfloat)-hw);
            glTexCoord2f(1, 1 - 0);
            glVertex3f((GLfloat)-hl, 0,           (GLfloat)-hw);

            glNormal3f(0, 0, 1);
	    glTexCoord2f(1, 1 - 0);
            glVertex3f((GLfloat)-hl, 0,           (GLfloat)-hw);
	    glTexCoord2f(0, 1 - 0);
            glVertex3f( (GLfloat)hl, 0,           (GLfloat)-hw);
	    glTexCoord2f(0, 1 - 1);
            glVertex3f( (GLfloat)hl, (GLfloat)hh, (GLfloat)-hw);
	    glTexCoord2f(1, 1 - 1);
            glVertex3f((GLfloat)-hl, (GLfloat)hh, (GLfloat)-hw);


	    /* Back side. */
            glNormal3f(0, 0, -1);
            glTexCoord2f(0, 1 - 1);
            glVertex3f((GLfloat)-hl, (GLfloat)hh, (GLfloat)hw);
            glTexCoord2f(1, 1 - 1);
            glVertex3f( (GLfloat)hl, (GLfloat)hh, (GLfloat)hw);
            glTexCoord2f(1, 1 - 0);
            glVertex3f( (GLfloat)hl, 0,           (GLfloat)hw);
            glTexCoord2f(0, 1 - 0);
            glVertex3f((GLfloat)-hl, 0,           (GLfloat)hw);

            glNormal3f(0, 0, 1);
	    glTexCoord2f(0, 1 - 0);
            glVertex3f((GLfloat)-hl, 0,           (GLfloat)hw);
	    glTexCoord2f(1, 1 - 0);
            glVertex3f( (GLfloat)hl, 0,           (GLfloat)hw);
	    glTexCoord2f(1, 1 - 1);
	    glVertex3f( (GLfloat)hl, (GLfloat)hh, (GLfloat)hw);
	    glTexCoord2f(0, 1 - 1);
	    glVertex3f((GLfloat)-hl, (GLfloat)hh, (GLfloat)hw);
	}
	glEnd();

	/* Ends. */
        V3DTextureSelect(end_tex);
	glBegin(GL_QUADS);
	{
	    /* Left end. */
            glNormal3f(-1, 0, 0);
	    glTexCoord2f(0, 1 - 0);
            glVertex3f((GLfloat)-hl, 0,           (GLfloat)-hw);
	    glTexCoord2f(1, 1 - 0);
            glVertex3f((GLfloat)-hl, 0,           (GLfloat)hw);
	    glTexCoord2f(1, 1 - 1);
            glVertex3f((GLfloat)-hl, (GLfloat)hh, (GLfloat)hw);
	    glTexCoord2f(0, 1 - 1);
            glVertex3f((GLfloat)-hl, (GLfloat)hh, (GLfloat)-hw);

            glNormal3f(1, 0, 0);
            glTexCoord2f(0, 1 - 1);
            glVertex3f((GLfloat)-hl, (GLfloat)hh, (GLfloat)-hw);
            glTexCoord2f(1, 1 - 1);
            glVertex3f((GLfloat)-hl, (GLfloat)hh, (GLfloat)hw);
            glTexCoord2f(1, 1 - 0);
            glVertex3f((GLfloat)-hl, 0,           (GLfloat)hw);
            glTexCoord2f(0, 1 - 0);
            glVertex3f((GLfloat)-hl, 0,           (GLfloat)-hw);


	    /* Right end. */
            glNormal3f(-1, 0, 0);
	    glTexCoord2f(0, 1 - 0);
            glVertex3f((GLfloat)hl, 0,           (GLfloat)-hw);
	    glTexCoord2f(1, 1 - 0);
            glVertex3f((GLfloat)hl, 0,           (GLfloat)hw);
	    glTexCoord2f(1, 1 - 1);
            glVertex3f((GLfloat)hl, (GLfloat)hh, (GLfloat)hw);
	    glTexCoord2f(0, 1 - 1);
            glVertex3f((GLfloat)hl, (GLfloat)hh, (GLfloat)-hw);

            glNormal3f(1, 0, 0);
            glTexCoord2f(0, 1 - 1);
            glVertex3f((GLfloat)hl, (GLfloat)hh, (GLfloat)-hw);
            glTexCoord2f(1, 1 - 1);
            glVertex3f((GLfloat)hl, (GLfloat)hh, (GLfloat)hw);
            glTexCoord2f(1, 1 - 0);
            glVertex3f((GLfloat)hl, 0,           (GLfloat)hw);
            glTexCoord2f(0, 1 - 0);
            glVertex3f((GLfloat)hl, 0,           (GLfloat)-hw);
	}
	glEnd();

	/* Other details, floatation strips. */
	V3DTextureSelect(NULL);
	glColor4f((GLfloat)0.9, (GLfloat)0.1, (GLfloat)0.1, (GLfloat)1.0);
	glBegin(GL_QUADS);
	{
            glNormal3f(0, 1, 0);
            glVertex3f((GLfloat)-hl,         (GLfloat)hh, (GLfloat)hw);
            glVertex3f((GLfloat)(-hl * 0.9), (GLfloat)hh, (GLfloat)hw);
            glVertex3f((GLfloat)(-hl * 0.9), (GLfloat)hh, (GLfloat)-hw);
            glVertex3f((GLfloat)-hl,         (GLfloat)hh, (GLfloat)-hw);

            glNormal3f(0, 1, 0);
            glVertex3f((GLfloat)hl,         (GLfloat)hh, (GLfloat)-hw);
            glVertex3f((GLfloat)(hl * 0.9), (GLfloat)hh, (GLfloat)-hw);
            glVertex3f((GLfloat)(hl * 0.9), (GLfloat)hh, (GLfloat)hw);
            glVertex3f((GLfloat)hl,         (GLfloat)hh, (GLfloat)hw);
        }
        glEnd();

	/* Begin drawing cables. */
        glBegin(GL_LINES);
        {
            double hl = l / 2;
            double hw = w / 2;
	    double hh = h / 2;

            glColor4f((GLfloat)0.9, (GLfloat)0.9, (GLfloat)0.9, (GLfloat)1.0);

            glNormal3f(0, 1, 0);
            glVertex3f((GLfloat)hl, (GLfloat)hh, (GLfloat)hw);
            glVertex3f(0, (GLfloat)h, 0);

            glVertex3f((GLfloat)hl, (GLfloat)hh, (GLfloat)-hw);
            glVertex3f(0, (GLfloat)h, 0);

            glVertex3f((GLfloat)-hl, (GLfloat)hh, (GLfloat)-hw);
            glVertex3f(0, (GLfloat)h, 0);

            glVertex3f((GLfloat)-hl, (GLfloat)hh, (GLfloat)hw);
            glVertex3f(0, (GLfloat)h, 0);

	    /* Cable up to hoist. */
	    glVertex3f(0, (GLfloat)h, 0);
	    glVertex3f(0, (GLfloat)rope_len, 0);
        }
        glEnd();

	return;
}

/*
 *	Draws smoke puff object.
 */
void SARDrawSmoke(
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_object_smoke_struct *obj_smoke_ptr
)
{
        int tex_num;
        v3d_texture_ref_struct *t;
	time_t life_span_max, life_span_cur;
        double r, life_span_coeff;


        /* Select texture. */
        tex_num = obj_smoke_ptr->tex_num;
        if(SARIsTextureAllocated(scene, tex_num))
            t = scene->texture_ref[tex_num];
        else
            return;

        /* Calculate life span coefficient from 1 to 4. */
	life_span_max = obj_ptr->life_span - obj_ptr->birth_time_ms;
	if(life_span_max <= 0)
	    return;

	life_span_cur = cur_millitime - obj_ptr->birth_time_ms;
	if(life_span_cur <= 0)
            return;

	life_span_coeff = 1.0 + ((double)life_span_cur /
	    (double)life_span_max * 3);

	/* Calculate current radius. */
	r = obj_smoke_ptr->radius * life_span_coeff;


        /* Select texture. */
        V3DTextureSelect(t);
        glColor4f(1.0, 1.0, 1.0, 1.0);

        glShadeModel(GL_FLAT);

        glBegin(GL_QUADS);
        {
            glNormal3f(0, 1, 0);

            glTexCoord2f(0, 1 - 0);
            glVertex3f((GLfloat)-r, (GLfloat)-r, 0);
            glTexCoord2f(1, 1 - 0);
            glVertex3f((GLfloat)r, (GLfloat)-r, 0);
            glTexCoord2f(1, 1 - 1);
            glVertex3f((GLfloat)r, (GLfloat)r, 0);
            glTexCoord2f(0, 1 - 1);
            glVertex3f((GLfloat)-r, (GLfloat)r, 0);
        }
        glEnd();

	return;
}

/*
 *	Draws explosion object.
 */
void SARDrawExplosion(
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_object_explosion_struct *obj_explosion_ptr
)
{
	int cur_frame, tex_num;
	v3d_texture_ref_struct *t;
	double r;


	/* Select current texture and frame. */
	cur_frame = obj_explosion_ptr->cur_frame;
	tex_num = obj_explosion_ptr->tex_num;
	if(SARIsTextureAllocated(scene, tex_num))
	    t = scene->texture_ref[tex_num];
	else
	    return;

	/* Sanitize frame number. */
	if(cur_frame >= t->total_frames)
	    cur_frame = t->total_frames - 1;
	if(cur_frame < 0)
	    cur_frame = 0;

	/* Get radius in meters. */
	r = obj_explosion_ptr->radius;

	/* Select texture. */
	V3DTextureSelectFrame(t, cur_frame);
        glColor4f(1.0, 1.0, 1.0, 1.0);

	glShadeModel(GL_FLAT);

	glBegin(GL_QUADS);
	{
	    glNormal3f(0, 1, 0);


            glTexCoord2f(0, 1 - 0);
            glVertex3f((GLfloat)-r, (GLfloat)-r, 0);
            glTexCoord2f(1, 1 - 0);
            glVertex3f((GLfloat)r, (GLfloat)-r, 0);
            glTexCoord2f(1, 1 - 1);
            glVertex3f((GLfloat)r, (GLfloat)r, 0);
            glTexCoord2f(0, 1 - 1);
            glVertex3f((GLfloat)-r, (GLfloat)r, 0);
	}
	glEnd();

	return;
}

/*
 *	Draws helipad.
 *
 *	All inputs assumed valid.
 */
void SARDrawHelipad(
	gw_display_struct *display, sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_object_helipad_struct *obj_helipad_ptr,
	double distance		/* 3d distance in meters. */
)
{
        int tex_num;
        v3d_texture_ref_struct *t;
	double x_min, x_max, y_min, y_max;


        /* Select current texture. */
        tex_num = obj_helipad_ptr->tex_num;
        if(SARIsTextureAllocated(scene, tex_num))
            t = scene->texture_ref[tex_num];
        else
            t = NULL;

        /* Get x and y bounds. */
	y_max = obj_helipad_ptr->length / 2;
        y_min = -y_max;
        x_max = obj_helipad_ptr->width / 2;
	x_min = -x_max;

        /* Select texture. */
	if(t == NULL)
	{
            V3DTextureSelect(NULL);
            glColor4f((GLfloat)0.6, (GLfloat)0.6, (GLfloat)0.6, (GLfloat)1.0);
	}
	else
	{
	    V3DTextureSelect(t);
            glColor4f((GLfloat)1.0, (GLfloat)1.0, (GLfloat)1.0, (GLfloat)1.0);
	}
        glShadeModel(GL_FLAT);

        glBegin(GL_QUADS);
        {
            glNormal3f(0, 1, 0);

            glTexCoord2f(0, 1 - 0);
            glVertex3f((GLfloat)x_min, 0, (GLfloat)-y_max);

            glTexCoord2f(0, 1 - 1);
            glVertex3f((GLfloat)x_min, 0, (GLfloat)-y_min);

            glTexCoord2f(1, 1 - 1);
            glVertex3f((GLfloat)x_max, 0, (GLfloat)-y_min);

            glTexCoord2f(1, 1 - 0);
            glVertex3f((GLfloat)x_max, 0, (GLfloat)-y_max);
        }
        glEnd();

	/* Unselect texture. */
	if(t != NULL)
            V3DTextureSelect(NULL);

        if(obj_helipad_ptr->flags & SAR_HELIPAD_FLAG_EDGE_LIGHTING)
        {
            STATE_FOR_POINT_LIGHTS_BEGIN

            StateGLEnable(&display->state_gl, GL_POINT_SMOOTH);
            StateGLPointSize(&display->state_gl, 1.0);
            glBegin(GL_POINTS);
            {
                glNormal3f(0, 1, 0);
                glColor4f(1.0, 1.0, 1.0, 1.0);

                glVertex3f((GLfloat)x_min, 0, (GLfloat)-y_min);
                glVertex3f((GLfloat)x_min, 0, (GLfloat)-y_max);
                glVertex3f((GLfloat)x_max, 0, (GLfloat)-y_max);
                glVertex3f((GLfloat)x_max, 0, (GLfloat)-y_min);
            }
            glEnd();

	    STATE_FOR_POINT_LIGHTS_END
	}

	return;
}

/*
 *	Draws runway.
 *
 *	All inputs assumed valid.
 */
void SARDrawRunway(
	gw_display_struct *display, sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_object_runway_struct *obj_runway_ptr,
	double distance		/* 3d distance in meters. */
)
{
        int tex_num;
        v3d_texture_ref_struct *t;
        double x_min, x_max, y_min, y_max, len;


        /* Select current texture. */
        tex_num = obj_runway_ptr->tex_num;
        if(SARIsTextureAllocated(scene, tex_num))
            t = scene->texture_ref[tex_num];
        else
            t = NULL;

        /* Get x and y bounds. */
	len = obj_runway_ptr->length;
        y_max = len / 2;
        y_min = -y_max;
        x_max = obj_runway_ptr->width / 2;
        x_min = -x_max;


	/* Farther than 5000 meters away? */
	if(distance > 5000)
	{
	    /* Draw simple. */

            V3DTextureSelect(NULL);
            glColor4f((GLfloat)0.3, (GLfloat)0.3, (GLfloat)0.3, (GLfloat)1.0);

            glShadeModel(GL_FLAT);

            glBegin(GL_QUADS);
            {
                glNormal3f(0, 1, 0);
 
                glVertex3f((GLfloat)x_min, 0, (GLfloat)-y_max);
                glVertex3f((GLfloat)x_min, 0, (GLfloat)-y_min);
                glVertex3f((GLfloat)x_max, 0, (GLfloat)-y_min);
                glVertex3f((GLfloat)x_max, 0, (GLfloat)-y_max);
            }
            glEnd();
	}
	else
	{
	    /* Draw closeup. */

            /* Select texture. */
            if(t == NULL)
            {
                V3DTextureSelect(NULL);
                glColor4f(0.5, 0.5, 0.5, 1.0);
            }
            else
            {
                V3DTextureSelect(t);
                glColor4f(1.0, 1.0, 1.0, 1.0);
            }       
            glShadeModel(GL_FLAT);

            glBegin(GL_QUADS);
            {
                glNormal3f(0, 1, 0);

                glTexCoord2f(0, 1 - 0);  
                glVertex3f((GLfloat)x_min, 0, (GLfloat)-y_max); 

                glTexCoord2f(0, 1 - 1);
                glVertex3f((GLfloat)x_min, 0, (GLfloat)-y_min);

                glTexCoord2f(1, 1 - 1);
                glVertex3f((GLfloat)x_max, 0, (GLfloat)-y_min);

                glTexCoord2f(1, 1 - 0);
                glVertex3f((GLfloat)x_max, 0, (GLfloat)-y_max);
            }
            glEnd();

	    if(t != NULL)
		V3DTextureSelect(NULL);

	    /* Dashes. */
	    glBegin(GL_QUADS);
	    {
                double dash_spacing, num_dashes, dash_width_min,
                       dash_width_max, dash_cur;

                num_dashes = 10;
                dash_spacing = MAX(len / (num_dashes * 2), 1.0);
                dash_width_min = -1.0;
                dash_width_max = 1.0;

                glColor4f(1.0, 1.0, 1.0, 1.0);
		glNormal3f(0, 1, 0);

		dash_cur = y_min;
		while((dash_cur + dash_spacing) <= y_max)
		{
		    glVertex3f(
			(GLfloat)dash_width_min, 0,
			(GLfloat)(-(dash_cur + dash_spacing))
		    );
                    glVertex3f(
			(GLfloat)dash_width_min, 0,
			(GLfloat)(-dash_cur)
		    );
                    glVertex3f(
			(GLfloat)dash_width_max, 0,
			(GLfloat)(-dash_cur)
		    );
                    glVertex3f(
			(GLfloat)dash_width_max, 0,
			(GLfloat)(-(dash_cur + dash_spacing))
		    );

		    dash_cur += (2 * dash_spacing);
		}
            }
            glEnd();
	}

	/* Draw edge lighting? */
	if(obj_runway_ptr->flags & SAR_RUNWAY_FLAG_EDGE_LIGHTING)
	{
	    double spacing;

	    STATE_FOR_POINT_LIGHTS_BEGIN

	    /* Calculate spacing of each light, min 100 meters apart. */
	    spacing = MAX(
		pow(2, floor(log(distance) / log(2.5))),
		100
	    );

	    glColor4f(1.0, 1.0, 1.0, 1.0);
            StateGLDisable(&display->state_gl, GL_POINT_SMOOTH);
            StateGLPointSize(&display->state_gl, 1.0);
            glBegin(GL_POINTS);
            {
		double t;

                glNormal3f(0, 1, 0);

	        for(t = 0; t < len; t += spacing)
	        {
                    glVertex3f((GLfloat)x_min, 0, (GLfloat)(-(y_min + t)));
                    glVertex3f((GLfloat)x_max, 0, (GLfloat)(-(y_min + t)));
                }

                glVertex3f((GLfloat)x_min, 0, (GLfloat)-y_max);
                glVertex3f((GLfloat)x_max, 0, (GLfloat)-y_max);
            }
	    glEnd();

	    STATE_FOR_POINT_LIGHTS_END
	}

	return;
}

/*
 *	Draws human.
 */
void SARDrawHuman(
	gw_display_struct *display, sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_object_human_struct *obj_human_ptr
)
{
	sar_position_struct *pos;
	sar_obj_flags_t flags;

	double anim_coeff = (double)obj_human_ptr->anim_pos /
	    (double)((sar_grad_anim_t)-1);

	double torso_angle;
	double left_bisep_angle, right_bisep_angle;
	double left_shoulder_angle, right_shoulder_angle;
	double left_trisep_angle, right_trisep_angle;
	double left_hip_to_thigh_angle, right_hip_to_thigh_angle;
	double left_knee_to_calv_angle, right_knee_to_calv_angle;

	double thigh_len = 0.3;
	double calv_len = 0.4;
	double torso_len = 0.8;
	double bisep_len = 0.3;
	double trisep_len = 0.4;
	double base_to_torso = 0;	/* Ground to bottom of torso,
					 * calculate later.
					 */

	pos = &obj_ptr->pos;
	flags = obj_human_ptr->flags;

	V3DTextureSelect(NULL);
        glShadeModel(GL_FLAT);


	/* Calculate angles of body parts. Note that all zero radians
	 * indicates human is standing upright with hands to side.
	 */
	if(flags & (SAR_HUMAN_FLAG_RUN_AWAY | SAR_HUMAN_FLAG_RUN_TOWARDS))
	{
            if(anim_coeff > 0.5)
                anim_coeff = (1.0 - anim_coeff) / 0.5;
            else
                anim_coeff = anim_coeff / 0.5;

            torso_angle = 0.1 * PI;
            left_shoulder_angle = 0.0 * PI;
            right_shoulder_angle = 0.0 * PI;
            left_bisep_angle = (1.7 + (0.35 * (1.0 - anim_coeff))) * PI;
            right_bisep_angle = (1.7 + (0.35 * anim_coeff)) * PI;
            left_trisep_angle = 1.8 * PI;
            right_trisep_angle = 1.8 * PI;
            left_hip_to_thigh_angle = (0.2 - (0.5 * anim_coeff)) * PI;
            right_hip_to_thigh_angle = (0.2 - (0.5 * (1.0 - anim_coeff))) * PI;
            left_knee_to_calv_angle = 0.2 * PI;
            right_knee_to_calv_angle = 0.2 * PI;

            base_to_torso = thigh_len + calv_len;
	}
	else if(flags & SAR_HUMAN_FLAG_LYING)
	{
            torso_angle = 1.5 * PI;
	    left_shoulder_angle = 0.0 * PI;
	    right_shoulder_angle = 0.0 * PI;
            left_bisep_angle = 0.0 * PI;
	    right_bisep_angle = 0.0 * PI;
            left_trisep_angle = 0.0 * PI;
            right_trisep_angle = 0.0 * PI;
            left_hip_to_thigh_angle = 0.0 * PI;
            right_hip_to_thigh_angle = 0.0 * PI;
            left_knee_to_calv_angle = 0.0 * PI;
            right_knee_to_calv_angle = 0.0 * PI;

            base_to_torso = thigh_len + calv_len;
	}
        else if(flags & SAR_HUMAN_FLAG_SIT)
        {
	    /* Sitting base at tush. */
            torso_angle = 0.0 * PI;
            left_shoulder_angle = 0.0 * PI;
            right_shoulder_angle = 0.0 * PI;
            left_bisep_angle = 0.0 * PI;
            right_bisep_angle = 0.0 * PI;
            left_trisep_angle = 1.75 * PI;
            right_trisep_angle = 1.75 * PI;
            left_hip_to_thigh_angle = 1.5 * PI;
            right_hip_to_thigh_angle = 1.5 * PI;
            left_knee_to_calv_angle = 0.5 * PI;
            right_knee_to_calv_angle = 0.5 * PI;

            base_to_torso = 0;
        }
        else if(flags & SAR_HUMAN_FLAG_SIT_DOWN)
        {
            torso_angle = 0.0 * PI;
            left_shoulder_angle = 0.0 * PI;
            right_shoulder_angle = 0.0 * PI;
            left_bisep_angle = 0.0 * PI;
            right_bisep_angle = 0.0 * PI;
            left_trisep_angle = 1.75 * PI;
            right_trisep_angle = 1.75 * PI;
            left_hip_to_thigh_angle = 1.5 * PI;
            right_hip_to_thigh_angle = 1.5 * PI;
            left_knee_to_calv_angle = 0.0 * PI;
            right_knee_to_calv_angle = 0.0 * PI;

            base_to_torso = 0;
	}
        else if(flags & SAR_HUMAN_FLAG_SIT_UP)
        {
	    /* Sitting base at feet. */
            torso_angle = 0.0 * PI;
            left_shoulder_angle = 0.0 * PI;
            right_shoulder_angle = 0.0 * PI;
            left_bisep_angle = 0.0 * PI;
            right_bisep_angle = 0.0 * PI;
            left_trisep_angle = 1.75 * PI;
            right_trisep_angle = 1.75 * PI;
            left_hip_to_thigh_angle = 1.5 * PI;
            right_hip_to_thigh_angle = 1.5 * PI; 
            left_knee_to_calv_angle = 0.5 * PI;
            right_knee_to_calv_angle = 0.5 * PI;
        
            base_to_torso = calv_len;
	}
	else
	{
	    /* Standing. */
	    if(flags & SAR_HUMAN_FLAG_IN_WATER)
	    {
		/* Standing in water, arms fore and moving. */
		if(anim_coeff > 0.5)
		    anim_coeff = (1.0 - anim_coeff) / 0.5;
		else
		    anim_coeff = anim_coeff / 0.5;

                torso_angle = 0.0 * PI;
                left_shoulder_angle = (0.4 * anim_coeff) * PI;
                right_shoulder_angle = (2 - (0.4 * anim_coeff)) * PI;
                left_bisep_angle = 1.5 * PI;
                right_bisep_angle = 1.5 * PI;
                left_trisep_angle = 0.0 * PI;
	        right_trisep_angle = 0.0 * PI;
                left_hip_to_thigh_angle = 0.0 * PI;
                right_hip_to_thigh_angle = 0.0 * PI;
                left_knee_to_calv_angle = 0.0 * PI;
                right_knee_to_calv_angle = 0.0 * PI;

		/* Since in water, move torso down. */
		base_to_torso = torso_len * -0.92;
	    }
	    else
	    {
		/* Standing on ground. */
		if((flags & SAR_HUMAN_FLAG_NEED_RESCUE) &&
		   (flags & SAR_HUMAN_FLAG_AWARE)
		)
		{
                 /* Need rescue and aware that rescuer is
		  * near by, move arms fore and wave.
		  */
                 if(anim_coeff > 0.5)
                    anim_coeff = (1.0 - anim_coeff) / 0.5;
                 else
                    anim_coeff = anim_coeff / 0.5;

                 torso_angle = 0.0 * PI;
                 left_shoulder_angle = (0.6 + (0.4 * anim_coeff)) * PI; 
                 right_shoulder_angle = (1.4 - (0.4 * anim_coeff)) * PI;
                 left_bisep_angle = 0.0 * PI; 
                 right_bisep_angle = 0.0 * PI; 
                 left_trisep_angle = 0.0 * PI; 
                 right_trisep_angle = 0.0 * PI;
                 left_hip_to_thigh_angle = 0.0 * PI; 
                 right_hip_to_thigh_angle = 0.0 * PI;  
                 left_knee_to_calv_angle = 0.0 * PI; 
                 right_knee_to_calv_angle = 0.0 * PI;

                 base_to_torso = thigh_len + calv_len;
		}
		else
		{
		 /* Just standing on ground. */
                 torso_angle = 0.0 * PI;
                 left_shoulder_angle = 0.0 * PI;
                 right_shoulder_angle = 0.0 * PI;
                 left_bisep_angle = 0.0 * PI;
                 right_bisep_angle = 0.0 * PI;
                 left_trisep_angle = 0.0 * PI;
                 right_trisep_angle = 0.0 * PI;
                 left_hip_to_thigh_angle = 0.0 * PI;
                 right_hip_to_thigh_angle = 0.0 * PI;
                 left_knee_to_calv_angle = 0.0 * PI;
                 right_knee_to_calv_angle = 0.0 * PI;

		 base_to_torso = thigh_len + calv_len;
		}
	    }
	}


	/* Begin drawing human. */

	/* Torso. */
        glPushMatrix();
        {
	    /* Move to base of torso (the tush ;) ). */
            glRotatef((GLfloat)-SFMRadiansToDegrees(torso_angle), 1.0, 0.0, 0.0);
            glTranslatef(0, (GLfloat)base_to_torso, 0);

	    /* Move to center of torso and draw torso. */
            glPushMatrix();
            {
		/* Hips of torso. */
                glBegin(GL_QUADS);
                {
                    glColor4f((GLfloat)0.2, (GLfloat)0.2, (GLfloat)0.2, (GLfloat)1.0);
	            SARVertexBoxBase(
			0.45 * torso_len,
			0.25 * torso_len,
			0.10 * torso_len
		    );
                }
                glEnd();

		glTranslatef(0, (GLfloat)(0.10 * torso_len), 0);

                /* Upper torso. */
                glBegin(GL_QUADS);
                {
                    glColor4f(1.0, 0.0, 0.0, 1.0);
                    SARVertexBoxBase(
                        0.45 * torso_len,
                        0.25 * torso_len,
                        0.90 * torso_len      
                    );
                }
                glEnd();
	    }
	    glPopMatrix();
	    /* Back to base of torso. */

	    /* Draw left leg. */
            glPushMatrix();
            {
		/* Thigh. */
		glRotatef((GLfloat)-SFMRadiansToDegrees(left_hip_to_thigh_angle), 1.0, 0.0, 0.0);
                glTranslatef((GLfloat)(-0.16 * torso_len), 0, 0);

                glBegin(GL_QUADS);
                {
                    glColor4f((GLfloat)0.8, (GLfloat)0.78, (GLfloat)0.75, (GLfloat)1.0);
                    SARVertexBoxBase(
                        (GLfloat)(0.5 * thigh_len),
                        (GLfloat)(0.5 * thigh_len),
                        (GLfloat)(-thigh_len)
                    );
                }
                glEnd();

		/* Calv. */
                glTranslatef(0, (GLfloat)-thigh_len, 0);
		glRotatef((GLfloat)-SFMRadiansToDegrees(left_knee_to_calv_angle), 1.0, 0.0, 0.0);

                glBegin(GL_QUADS);
                {
                    glColor4f((GLfloat)0.8, (GLfloat)0.78, (GLfloat)0.75, (GLfloat)1.0);
                    SARVertexBoxBase(
                        (GLfloat)(0.3 * calv_len),
                        (GLfloat)(0.3 * calv_len),
                        (GLfloat)-calv_len
                    );
                }
                glEnd();

		/* Feet. */
                glTranslatef(0, (GLfloat)-calv_len, (GLfloat)-0.05);
                glBegin(GL_QUADS);
                {
                    glColor4f((GLfloat)0.3, (GLfloat)0.3, (GLfloat)0.3, (GLfloat)1.0);
                    SARVertexBoxBase(
                        0.1,
                        0.2,
                        -0.1
                    );
                }
                glEnd();
	    }
            glPopMatrix();
            /* Back to base of torso. */

            /* Draw right leg. */
            glPushMatrix();
            {
                /* Thigh. */
		glRotatef((GLfloat)-SFMRadiansToDegrees(right_hip_to_thigh_angle), 1.0, 0.0, 0.0);
                glTranslatef((GLfloat)(0.16 * torso_len), 0, 0);

                glBegin(GL_QUADS);
                {
                    glColor4f((GLfloat)0.8, (GLfloat)0.78, (GLfloat)0.75, (GLfloat)1.0);
                    SARVertexBoxBase(
                        (GLfloat)(0.5 * thigh_len),
                        (GLfloat)(0.5 * thigh_len),
                        (GLfloat)-thigh_len
                    );  
                }
                glEnd();   
                
                /* Calv. */
                glTranslatef(0, (GLfloat)-thigh_len, 0);
		glRotatef((GLfloat)-SFMRadiansToDegrees(right_knee_to_calv_angle), 1.0, 0.0, 0.0);

                glBegin(GL_QUADS);
                {
                    glColor4f((GLfloat)0.8, (GLfloat)0.78, (GLfloat)0.75, (GLfloat)1.0);
                    SARVertexBoxBase(
                        (GLfloat)(0.3 * calv_len),
                        (GLfloat)(0.3 * calv_len),
                        (GLfloat)-calv_len
                    );
                }
                glEnd();

                /* Feet. */
                glTranslatef(0, (GLfloat)-calv_len, (GLfloat)-0.05);
                glBegin(GL_QUADS);
                {
                    glColor4f((GLfloat)0.3, (GLfloat)0.3, (GLfloat)0.3, (GLfloat)1.0);
                    SARVertexBoxBase( 
                        0.1,
                        0.2,
                        -0.1
                    );
                }
                glEnd();
            }
            glPopMatrix();
            /* Back to base of torso. */

	    /* Move to shoulders. */
	    glTranslatef(0, (GLfloat)(0.95 * torso_len), 0);

            /* Draw left arm. */
            glPushMatrix();
            {
                /* Bisep and shoulder. */
		glTranslatef(
		    (GLfloat)((-0.225 * torso_len) + (-0.25 * bisep_len)),
		    0,
		    0
		);
                glRotatef((GLfloat)-SFMRadiansToDegrees(left_bisep_angle), 1.0, 0.0, 0.0);
		glRotatef((GLfloat)-SFMRadiansToDegrees(left_shoulder_angle), 0.0, 0.0, 1.0);

                glBegin(GL_QUADS);
                {
                    glColor4f((GLfloat)0.2, (GLfloat)0.2, (GLfloat)1.0, (GLfloat)1.0);
                    SARVertexBoxBase(
                        0.5 * bisep_len,
                        0.5 * bisep_len,
                        -bisep_len
                    );
                }
                glEnd();

                /* Trisep. */
                glTranslatef(0, (GLfloat)-bisep_len, 0);
		glRotatef((GLfloat)-SFMRadiansToDegrees(left_trisep_angle), 1.0, 0.0, 0.0);

                glBegin(GL_QUADS);
                {
                    glColor4f((GLfloat)0.8, (GLfloat)0.78, (GLfloat)0.75, (GLfloat)1.0);
                    SARVertexBoxBase(
                        0.28 * trisep_len,
                        0.28 * trisep_len,
                        -trisep_len
                    );
                }
                glEnd();

                /* Hand. */
                glTranslatef(0, (GLfloat)-trisep_len, 0);
                glBegin(GL_QUADS);
                {
                    glColor4f((GLfloat)1.0, (GLfloat)0.8, (GLfloat)0.4, (GLfloat)1.0);
                    SARVertexBoxBase(0.2, 0.1, -0.2);
                }
                glEnd();
	    }
            glPopMatrix();
            /* Back to shoulders. */

	    /* Draw right arm. */
            glPushMatrix();
            {
                /* Bisep. */
                glTranslatef(
		    (GLfloat)((0.225 * torso_len) + (0.25 * bisep_len)),
		    0,
		    0
		);
		glRotatef((GLfloat)-SFMRadiansToDegrees(right_bisep_angle), 1.0, 0.0, 0.0);
		glRotatef((GLfloat)-SFMRadiansToDegrees(right_shoulder_angle), 0.0, 0.0, 1.0);

                glBegin(GL_QUADS);
                {
                    glColor4f((GLfloat)0.2, (GLfloat)0.2, (GLfloat)1.0, (GLfloat)1.0);
                    SARVertexBoxBase(
                        0.5 * bisep_len,
                        0.5 * bisep_len,
                        -bisep_len
                    );
                }
                glEnd();

                /* Trisep. */
                glTranslatef(0, (GLfloat)-bisep_len, 0);
		glRotatef((GLfloat)-SFMRadiansToDegrees(right_trisep_angle), 1.0, 0.0, 0.0);

                glBegin(GL_QUADS);
                {
                    glColor4f((GLfloat)0.8, (GLfloat)0.78, (GLfloat)0.75, (GLfloat)1.0);
                    SARVertexBoxBase(
                        (GLfloat)(0.28 * trisep_len),
                        (GLfloat)(0.28 * trisep_len),
                        (GLfloat)-trisep_len
                    );
                }
                glEnd();

		/* Hand. */
                glTranslatef(0, (GLfloat)-trisep_len, 0);
		glBegin(GL_QUADS);
                {
                    glColor4f((GLfloat)1.0, (GLfloat)0.8, (GLfloat)0.4, (GLfloat)1.0);
                    SARVertexBoxBase(0.2, 0.1, -0.2);
                }
                glEnd();
            }
            glPopMatrix();
            /* Back to shoulders. */

            /* Head. */
            glTranslatef(0, (GLfloat)(0.05 * torso_len), 0);
            glBegin(GL_QUADS);
            {
                glColor4f((GLfloat)1.0, (GLfloat)0.8, (GLfloat)0.4, (GLfloat)1.0);
                SARVertexBoxBase(0.18, 0.18, 0.25);
            }
            glEnd();

	    /* Hair. */
	    glPushMatrix();
	    {
                glTranslatef(0, 0.25, 0);
                glBegin(GL_QUADS);
                {
                    glColor4f((GLfloat)0.25, (GLfloat)0.23, (GLfloat)0.20, (GLfloat)1.0);
                    SARVertexBoxBase(0.18, 0.18, 0.1);
                }
                glEnd();
	    }
	    glPopMatrix();
            glPushMatrix();
            {
                glTranslatef((GLfloat)0.0, (GLfloat)0.05, (GLfloat)(-(-0.09 - 0.03)));
                glBegin(GL_QUADS);
                {
                    glColor4f((GLfloat)0.25, (GLfloat)0.23, (GLfloat)0.20, (GLfloat)1.0);
                    SARVertexBoxBase(0.18, 0.06, 0.25);
                }
                glEnd();
            }
            glPopMatrix();


        }
        glPopMatrix();

	/* Draw water ripples if in water. */
	if(flags & SAR_HUMAN_FLAG_IN_WATER)
	{
	    GLboolean depth_mask_flag;
	    int tex_num, frame_num;
	    v3d_texture_ref_struct *t;
	    double hr = 5.0;	/* Half ripple radius. */


	    tex_num = obj_human_ptr->water_ripple_tex_num;
            if(SARIsTextureAllocated(scene, tex_num) &&
	       option.textured_objects
	    )
	    {
		t = scene->texture_ref[tex_num];

		/* Select frame by animation, re-update anim_coeff. */
		anim_coeff = (double)obj_human_ptr->anim_pos /
                    (double)((sar_grad_anim_t)-1);

		frame_num = (int)(anim_coeff * t->total_frames);
		if(frame_num >= t->total_frames)
		    frame_num = t->total_frames - 1;
		if(frame_num < 0)
		    frame_num = 0;
		V3DTextureSelectFrame(t, frame_num);

		glColor4f(1.0, 1.0, 1.0, 0.5);

                StateGLEnable(&display->state_gl, GL_BLEND);
                StateGLDisable(&display->state_gl, GL_ALPHA_TEST);
                depth_mask_flag = display->state_gl.depth_mask_flag;
                StateGLDepthMask(&display->state_gl, GL_FALSE);
		StateGLEnable(&display->state_gl, GL_POLYGON_OFFSET_FILL);
		StateGLBlendFunc(
		    &display->state_gl, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA
		);
		StateGLPolygonOffset(
		    &display->state_gl,
		    (GLfloat)option.gl_polygon_offset_factor, -1.0
		);

		glBegin(GL_QUADS);
		{
		    glNormal3f(0, 1, 0);
                    glTexCoord2f(0, 1 - 0);
                    glVertex3f((GLfloat)-hr, 0, (GLfloat)hr);
                    glTexCoord2f(1, 1 - 0);
                    glVertex3f((GLfloat)hr, 0, (GLfloat)hr);
                    glTexCoord2f(1, 1 - 1);
                    glVertex3f((GLfloat)hr, 0, (GLfloat)-hr);
                    glTexCoord2f(0, 1 - 1);
                    glVertex3f((GLfloat)-hr, 0, (GLfloat)-hr);
		}
		glEnd();

                StateGLDisable(&display->state_gl, GL_POLYGON_OFFSET_FILL);
                StateGLDisable(&display->state_gl, GL_BLEND);
		StateGLDepthMask(&display->state_gl, depth_mask_flag);
                StateGLEnable(&display->state_gl, GL_ALPHA_TEST);
	    }
	}

	return;
}


/*
 *	Draws HUD attitude ticks.
 *
 *      Inputs are assumed valid.
 */
void SARDrawHUDAttitude(
	gw_display_struct *display, sar_scene_struct *scene,
	sar_direction_struct *dir, sar_position_struct *vel,
	int offset_x, int offset_y,
	int width, int height,
	double fovx, double fovy
)
{
        int x, y, x2, y2, degi;
	double pitch_deg;
        int half_width, half_height;
	double sin_bank, cos_bank, r;
        int tick_start_y, end_min, end_max;	/* In pixels. */
	int	tick_inc_yi,			/* In pixels. */
		tick_inc_degreesi,		/* In degrees. */
		tick_x_max,			/* In pixels. */
		tick_x_min;
	double	tick_inc_y,			/* In pixels. */
		tick_inc_degrees;		/* In degrees. */
	int tick_start_pitch;			/* In degrees. */
	char text[80];

	/* Glide slope cursor. */
	int gs_w = 15, gs_h = 6;
	u_int8_t gs_bm[] = {
 0x03, 0x80,
 0x04, 0x40,
 0xfc, 0x7e,
 0x04, 0x40,
 0x03, 0x80,
 0x01, 0x00
	};

	/* Coordinates are handled in left hand rule xy plane,
	 * the y axis is inverted at the call to the X draw
	 * function.
	 */

        /* Calculate half of size. */
        half_width = width / 2;
        half_height = height / 2;

        /* Calculate limits and increments. */
        end_min = (int)(-0.5 * half_height);
        end_max = (int)(0.5 * half_height);

        tick_inc_degrees = 10;		/* Always 10 degrees. */

	tick_x_max = (int)(0.06 * width);
	tick_x_min = (int)(0.03 * width);

	sin_bank = sin(dir->bank);
	cos_bank = cos(dir->bank);

	/* Calculate tick_inc_y. */
	r = (double)height / fovy;
	tick_inc_y = r * SFMDegreesToRadians(tick_inc_degrees);

	/* Split into two cases, positive and negative pitch. */
	if(dir->pitch > PI)
	{
	    /* Positive pitch. */
	    double deg;

            pitch_deg = 360 - (int)SFMRadiansToDegrees(dir->pitch);
            deg = (int)pitch_deg % (int)tick_inc_degrees;
            tick_start_y = (int)(-deg * tick_inc_y / tick_inc_degrees);
            tick_start_pitch = (int)(pitch_deg - deg);
	}
	else
	{
	    /* Negative pitch. */
	    double deg;

            pitch_deg = (int)SFMRadiansToDegrees(dir->pitch);
            deg = tick_inc_degrees -
		((int)pitch_deg % (int)tick_inc_degrees);
            tick_start_y = (int)(-deg * tick_inc_y / tick_inc_degrees);
            tick_start_pitch = (int)(-pitch_deg - deg);
	}
	tick_inc_degreesi = (int)tick_inc_degrees;
	tick_inc_yi = (int)tick_inc_y;


/* Define procedure to draw each tick for use in the for() loop
 * just below.
 */
#define DO_DRAW_ATTITUDE_TICK	{ \
 x2 = (int)(-sin_bank * y); \
 y2 = (int)(cos_bank * y); \
 if(degi == 0) \
 { \
   glBegin(GL_LINES); \
   { \
     glVertex2f( \
	(GLfloat)(offset_x + x2 + (tick_x_min * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 + (tick_x_min * sin_bank) + half_height) \
     ); \
     glVertex2f( \
	(GLfloat)(offset_x + x2 + ((8 + tick_x_max) * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 + ((8 + tick_x_max) * sin_bank) + half_height) \
     ); \
  \
     glVertex2f( \
	(GLfloat)(offset_x + x2 - (tick_x_min * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 - (tick_x_min * sin_bank) + half_height) \
     ); \
     glVertex2f( \
	(GLfloat)(offset_x + x2 - ((8 + tick_x_max) * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 - ((8 + tick_x_max) * sin_bank) + half_height) \
     ); \
   } \
   glEnd(); \
 } \
 else \
 { \
   glBegin(GL_LINES); \
   { \
     glVertex2f( \
	(GLfloat)(offset_x + x2 + (tick_x_min * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 + (tick_x_min * sin_bank) + half_height) \
     ); \
     glVertex2f( \
	(GLfloat)(offset_x + x2 + (tick_x_max * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 + (tick_x_max * sin_bank) + half_height) \
     ); \
 \
     glVertex2f( \
	(GLfloat)(offset_x + x2 - (tick_x_min * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 - (tick_x_min * sin_bank) + half_height) \
     ); \
     glVertex2f( \
	(GLfloat)(offset_x + x2 - (tick_x_max * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 - (tick_x_max * sin_bank) + half_height) \
     ); \
  } \
  glEnd(); \
  \
    sprintf(text, "%i", degi); \
    /* Right. */ \
    GWDrawString( \
	display, \
	(int)(offset_x + x2 + (tick_x_max * cos_bank) + half_width \
	    - 8 + (12 * cos_bank)), \
	(int)(height - (offset_y + y2 + (tick_x_max * sin_bank) + \
	    half_height + 3)), \
        text \
    ); \
    /* Left. */ \
    GWDrawString( \
        (gw_display_struct *)display, \
        (int)(offset_x + x2 - (tick_x_max * cos_bank) + half_width \
            - 8 - (18 * cos_bank)), \
        (int)(height - (offset_y + y2 - (tick_x_max * sin_bank) + \
            half_height + 3)), \
        text \
    ); \
 } \
 degi += tick_inc_degreesi; \
}


        degi = tick_start_pitch;

	/* Draw upper (pitch up) ticks. */
        for(y = tick_start_y; y < end_max; y += tick_inc_yi)
	{
	    if(degi > 90)
                break;

	    DO_DRAW_ATTITUDE_TICK
	}

        degi = tick_start_pitch - tick_inc_degreesi;

        /* Draw lower (pitch down) ticks. */
        for(y = tick_start_y - tick_inc_yi; y > end_min; y -= tick_inc_yi)
	{
	    if(degi < -90)
		break;

	    DO_DRAW_ATTITUDE_TICK
	}
#undef DO_DRAW_ATTITUDE_TICK


	/* Draw nose axis marker. */
	x = offset_x + half_width;
	y = offset_y + half_height;
	glBegin(GL_LINES);
	{
	    glVertex2f(
		(GLfloat)(x - (0.05 * width)),
		(GLfloat)y
	    );
	    glVertex2f(
		(GLfloat)(x - (0.02 * width)),
		(GLfloat)y
	    );

	    glVertex2f(
		(GLfloat)(x + (0.05 * width)),
		(GLfloat)y
	    );
	    glVertex2f(
		(GLfloat)(x + (0.02 * width)),
		(GLfloat)y
	    );
	}
	glEnd();

	/* Draw glide slope marker. */
	if(vel != NULL)
	{
	    double h, theta;

	    /* Calculate angle to ground relative to velocity. */
	    theta = SFMSanitizeRadians(
		(2 * PI) - atan2(vel->z, vel->y)
	    );
	    /* Adjust angle to ground relative to aircraft's pitch. */
	    theta = SFMSanitizeRadians(
		theta - dir->pitch
	    );
	    if(theta > (0.5 * PI))
	    {
/* Don't go up beyond nose level. */
		h = 0;
		theta = 0;
	    }
	    else
	    {
	        h = (height / fovy) * theta;
	    }

	    if(h < -end_min)
	    {
                x = (int)(offset_x + half_width);
                y = (int)(offset_y + half_height - h);

		glRasterPos2i(x - (gs_w / 2), y - (gs_h / 2));
		glBitmap(
                    gs_w, gs_h,
                    0.0, 0.0,
                    (GLfloat)gs_w, 0.0,
		    gs_bm
                );
	    }
	}

	return;
}

/*
 *	Draws HUD heading ticks and intercept arrow.
 *
 *	Inputs are assumed valid.
 */
void SARDrawHUDHeading(
	gw_display_struct *display,
	sar_scene_struct *scene,
	sar_direction_struct *dir,
	sar_obj_intercept_struct *intercept,
	int offset_x, int offset_y,	/* Look offset. */
	int width, int height		/* Of window. */
)
{
	int x, y, deg, heading_deg, len;
	int half_width, half_height;
	int tick_start_x, end_min, end_max;	/* In pixels. */
	int	tick_inc_x,			/* In pixels. */
		tick_inc_degrees;		/* In degrees. */
	int tick_start_heading;			/* In degrees. */
	char text[256];
	int ia_w = 8, ia_h = 4;
	u_int8_t ia_bm[] = {0xc6, 0x6c, 0x38, 0x10};


        /* Coordinates are handled in left hand rule xy plane,
         * with the y axis already inverted!
         */

	/* Calculate half of size. */
	half_width = width / 2;
	half_height = height / 2;

	/* Calculate limits and increments. */
	end_min = (int)(half_width - (0.3 * half_width));
	end_max = (int)(half_width + (0.3 * half_width));

	tick_inc_x = (int)(0.09 * width);
	tick_inc_degrees = 10;		/* Always 10 degrees. */

	heading_deg = (int)SFMRadiansToDegrees(dir->heading);
	deg = heading_deg % tick_inc_degrees;
	tick_start_x = half_width - (deg * tick_inc_x / tick_inc_degrees);
	tick_start_heading = (int)SFMSanitizeDegrees(
	    heading_deg - deg
	);


#define DO_DRAW_HEADING_TICK	{ \
 glBegin(GL_LINES); \
 { \
  glVertex2f( \
     (GLfloat)(x + offset_x), \
     (GLfloat)(height - (y + offset_y - 10)) \
  ); \
  glVertex2f( \
     (GLfloat)(x + offset_x), \
     (GLfloat)(height - (y + offset_y - 14)) \
  ); \
 } \
 glEnd(); \
  \
 sprintf(text, "%i", deg / 10); \
 len = strlen(text); \
 GWDrawString( \
     display, \
     (int)(x + offset_x - ((len - 1) * 4) - 3), \
     (int)(y + offset_y - 11), \
     text \
 ); \
 deg = (int)SFMSanitizeDegrees(deg + tick_inc_degrees); \
}

	y = (int)(half_height - (0.35 * height));
	deg = tick_start_heading;
	for(x = tick_start_x; x < end_max; x += tick_inc_x)
	{
	    DO_DRAW_HEADING_TICK
	}

	deg = (int)SFMSanitizeDegrees(tick_start_heading - tick_inc_degrees);
        for(x = tick_start_x - tick_inc_x; x > end_min; x -= tick_inc_x)
        {
            DO_DRAW_HEADING_TICK
        }

#undef DO_DRAW_HEADING_TICK

	/* Draw intercept tick. */
	if(intercept != NULL)
	{
	    int intercept_tick_dx;
	    double dtheta, icpt_dist;
            double	icpt_dx = intercept->x - camera_pos.x,
			icpt_dy = intercept->y - camera_pos.y;


            /* Calculate intercept distance in meters. */
            icpt_dist = hypot(icpt_dx, icpt_dy);

	    /* Calculate delta theta from current heading to
	     * intercept bearing.
	     */
	    dtheta = SFMDeltaRadians(
		dir->heading,
		SFMSanitizeRadians(
		    (0.5 * PI) - atan2(icpt_dy, icpt_dx)
		)
	    );
	    /* Convert to degrees, no sanitize. */
	    dtheta = dtheta * (180 / PI);

	    if(tick_inc_degrees > 0)
	        intercept_tick_dx = (int)(dtheta * ((double)tick_inc_x /
		    (double)tick_inc_degrees));
	    else
		intercept_tick_dx = 0;

            /* Apply half width. */
            intercept_tick_dx += half_width;

	    if(intercept_tick_dx < end_min)
		intercept_tick_dx = end_min;
	    else if(intercept_tick_dx > end_max)
		intercept_tick_dx = end_max;

            glRasterPos2i(
		intercept_tick_dx + offset_x + 1 - (ia_w / 2),
		height - (y + offset_y + 5 - (ia_h / 2))
	    );
            glBitmap(
                ia_w, ia_h,
                0.0, 0.0,
                (GLfloat)ia_w, 0.0,
                ia_bm
            );

	    /* Draw distance to intercept. */
	    if(1)
	    {
		double v1 = SFMMetersToMiles(icpt_dist);
		if(v1 > 10.0)
		    sprintf(text, "R: %.0f", v1);
		else
		    sprintf(text, "R: %.1f", v1);

		GWDrawString(
		    display,
		    (int)(offset_x + half_width + (0.4 * half_width)),
		    (int)(offset_y + y),
		    text
		);
	    }
	}

	/* Current heading tick. */
	glBegin(GL_LINES);
	{
	    glVertex2f(
		(GLfloat)(half_width + offset_x),
		(GLfloat)(height - (y + offset_y + 1))
	    );
	    glVertex2f(
		(GLfloat)half_width + offset_x,
		(GLfloat)(height - (y + offset_y + 5))
	    );
	}

	return;
}



/*
 *	Draw HUD.
 *
 *	The given aspect is the width of the window in pixels
 *	divided by the height of the window in pixels.
 *
 *	All units are in projected units and radians, projected
 *	units are derived from the given width and height.
 */
void SARDrawHUD(
	gw_display_struct *display,
	sar_scene_struct *scene,
	sar_object_struct ***ptr,
	int *total,
	double aspect
)
{
	int i;
	sar_object_struct *obj_ptr;
        sar_direction_struct *cam_dir, *dir;
        sar_position_struct *pos, *vel = NULL;
	sar_object_aircraft_struct *obj_aircraft_ptr = NULL;
	sar_obj_intercept_struct *intercept = NULL;
	double fovx, fovz;	/* Field of view x and z, in radians. */
	int engine_state = SAR_ENGINE_STATE_OFF;
	double cockpit_offset_z = 0;
	double speed = 0;	/* In meters per cycke. */
	double throttle = 0;
	int gear_state = 0;
	double fuel = 0;
	sar_color_struct *color;
	GWFont *font;
	char text[80];


	/* Get initial values. */
	obj_ptr = scene->player_obj_ptr;
	if(obj_ptr == NULL)
	    return;
	if(obj_ptr->type == SAR_OBJ_TYPE_AIRCRAFT)
	    obj_aircraft_ptr = obj_ptr->data;

	if(aspect <= 0)
	    return;

        fovz = scene->camera_fovz;
	if(fovz <= 0)
	    return;
        fovx = fovz * aspect;

	color = &option.hud_color;
	font = option.hud_font;


        /* Get direction of camera in the cockpit. */
        cam_dir = &scene->camera_cockpit_dir;

	/* Get direction and position of object. */
	dir = &obj_ptr->dir;
        pos = &obj_ptr->pos;

	/* Get other values. */
	if(obj_aircraft_ptr != NULL)
	{
	    cockpit_offset_z = obj_aircraft_ptr->cockpit_offset_pos.z;
	    vel = &obj_aircraft_ptr->vel;
	    speed = obj_aircraft_ptr->speed;
	    engine_state = obj_aircraft_ptr->engine_state;
	    throttle = obj_aircraft_ptr->throttle;
	    fuel = obj_aircraft_ptr->fuel;

	    i = obj_aircraft_ptr->cur_intercept;
	    if((i >= 0) && (i < obj_aircraft_ptr->total_intercepts))
		intercept = obj_aircraft_ptr->intercept[i];

	    if(obj_aircraft_ptr->total_landing_gears > 0)
		gear_state = ((obj_aircraft_ptr->landing_gear[0] == NULL) ?
		    0 :
		    (obj_aircraft_ptr->landing_gear[0]->flags & SAR_LGEAR_FLAG_STATE)
		);
	}


	/* Check if camera direction is looking forward enough. */
	if(((cam_dir->heading > (1.5 * PI)) ||
            (cam_dir->heading < (0.5 * PI))) &&
           ((cam_dir->pitch > (1.5 * PI)) ||
            (cam_dir->pitch < (0.5 * PI)))
	)
	{
	    double heading, pitch;
	    double r;
	    int x, y;			/* Center of hud in pixels. */
	    int win_width, win_height;


	    heading = cam_dir->heading;
            if(heading > PI)
                heading = heading - (2 * PI);

	    pitch = cam_dir->pitch;
            if(pitch > PI)
                pitch = pitch - (2 * PI);

	    win_width = display->width;
	    win_height = display->height;

	    r = win_width / fovx;
	    x = (int)((win_width / 2) - (r * heading));

            r = win_height / fovz;
            y = (int)((win_height / 2) - (r * pitch));

	    sprintf(text, "%.1f", SFMMPCToMPH(speed));

	    SARDrawSetColor(color);
	    GWSetFont(display, font);
	    /* Speed in miles per hour. */
	    GWDrawString(
		display,
		(int)(x - (win_width * 0.20) - 25),
		(int)(y + 3),
		text
	    );

	    /* MSL altitude (include scene MSL elevation). */
            sprintf(
		text, "%.0f",
                SFMMetersToFeet(pos->z + cockpit_offset_z + scene->msl_elevation)
	    );
            GWDrawString(
                display, 
                (int)(x + (win_width * 0.20)),
                (int)(y + 3),
                text
            );

	    /* Throttle. */
	    switch(engine_state)
	    {
	      case SAR_ENGINE_STATE_ON:
		sprintf(text, "Thr: %.0f%%", throttle * 100);
		break;

	      case SAR_ENGINE_STATE_INIT:
		sprintf(text, "Eng(I): %.0f%%", throttle * 100);
                break;

	      case SAR_ENGINE_STATE_OFF:
                sprintf(text, "Eng(O): %.0f%%", throttle * 100);
                break;
	    }
            GWDrawString(
                display,
                (int)(x - (win_width * 0.25)),
                (int)(y + (win_height * 0.29)),
                text
            );

	    /* Air brakes. */
	    if(player_air_brakes)
	    {
                GWDrawString(
                    display,
                    (int)(x - (win_width * 0.25)),
                    (int)(y + (win_height * 0.29) - 16),
                    "Brakes"
                );
	    }


            /* Gear state. */
	    if(!gear_state)
	    {
                GWDrawString(
                    display,
                    (int)(x - (win_width * 0.25)),
                    (int)(y + (win_height * 0.29) + 16),
                    "Gear"
                );
	    }

	    /* Fuel. */
            sprintf(text, "Fuel: %.0f lbs", SFMKGToLBS(fuel));
	    GWDrawString(
                display,
                (int)(x + (win_width * 0.05)),
                (int)(y + (win_height * 0.29)),
                text  
            );

	    /* Draw HUD heading ticks. */
	    SARDrawHUDHeading(
	        display,
	        scene,
	        dir,
		intercept,
		(int)(r * -heading), (int)(r * -pitch),
		win_width, win_height
	    );

	    /* Draw HUD attitude ticks. */
	    SARDrawHUDAttitude(
		display, scene,
                dir, vel,
                (int)(r * -heading), (int)(r * pitch),
                win_width, win_height,
		fovx, fovz
	    );
	}

	return;
}

/*
 *	Draws attitude and position label strip of the specified
 *	object.
 */
void SARDrawOutsideAttitude(
	gw_display_struct *display,
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
        int width, int height
)
{
        int i, x, y, x2, y2, degi, offset_x, offset_y;
        double pitch_deg;
        int half_width, half_height;
        double sin_bank, cos_bank;
        int tick_start_y, end_min, end_max;     /* In pixels. */
        int     tick_inc_yi,                    /* In pixels. */
                tick_inc_degreesi,              /* In degrees. */
                tick_x_max,                     /* In pixels. */
                tick_x_min;
        double  tick_inc_y,                     /* In pixels. */
                tick_inc_degrees;               /* In degrees. */
        int tick_start_pitch;                   /* In degrees. */
	sar_position_struct *pos;
	sar_direction_struct *dir;
	sar_object_aircraft_struct *obj_aircraft_ptr = NULL;
        char text[128];

	int vel_dir_w = 32, vel_dir_h = 32;
	u_int8_t vel_dir_bm[] = {
0x00, 0x00, 0x00, 0x00,
0x00, 0x01, 0x00, 0x00, 
0x00, 0x01, 0x00, 0x00, 
0x10, 0x00, 0x00, 0x08, 
0x08, 0x00, 0x00, 0x10, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x01, 0x00, 0x00, 
0xc0, 0x03, 0x80, 0x03, 
0x00, 0x01, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x00, 0x00, 0x00, 0x00, 
0x08, 0x00, 0x00, 0x10, 
0x10, 0x00, 0x00, 0x08, 
0x00, 0x01, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00
	};


        /* Coordinates are handled in left hand rule xy plane,
         * the y axis is inverted at the call to the X draw
         * function.
         */

	if(obj_ptr == NULL)
	    return;

	pos = &obj_ptr->pos;
	dir = &obj_ptr->dir;

	switch(obj_ptr->type)
	{
	  case SAR_OBJ_TYPE_AIRCRAFT:
	    obj_aircraft_ptr = obj_ptr->data;
	    break;
	}

        /* Calculate half of size. */
        half_width = width / 2;   
        half_height = height / 2;

	offset_x = 0;
	offset_y = -half_height + 20 + 5;

        /* Calculate Y limits and increments. */
        end_min = -20;
        end_max = 20;

        tick_inc_degrees = 10;          /* Always 10 degrees. */

        tick_x_max = 10;		/* Tick width. */
        tick_x_min = 4;			/* Tick spacing from center. */

        sin_bank = sin(dir->bank);
        cos_bank = cos(dir->bank);

        /* Set tick increment along Y axis. */
        tick_inc_y = 10;

        /* Split into two cases, positive and negative pitch. */ 
        if(dir->pitch > PI)
        {
            /* Positive pitch. */
            double deg;

            pitch_deg = 360 - (int)SFMRadiansToDegrees(dir->pitch);
            deg = (int)pitch_deg % (int)tick_inc_degrees;
            tick_start_y = (int)(-deg * tick_inc_y / tick_inc_degrees);
            tick_start_pitch = (int)(pitch_deg - deg);
        }
        else
        {       
            /* Negative pitch. */
            double deg;

            pitch_deg = (int)SFMRadiansToDegrees(dir->pitch);
            deg = tick_inc_degrees -
                ((int)pitch_deg % (int)tick_inc_degrees);
            tick_start_y = (int)(-deg * tick_inc_y / tick_inc_degrees);
            tick_start_pitch = (int)(-pitch_deg - deg);
        }   
        tick_inc_degreesi = (int)tick_inc_degrees;
        tick_inc_yi = (int)tick_inc_y; 


/* Define procedure to draw each tick for use in the for() loop
 * just below.
 */
#define DO_DRAW_ATTITUDE_TICK	{ \
 x2 = (int)(-sin_bank * y); \
 y2 = (int)(cos_bank * y); \
 if(degi == 0) \
 { \
    glBegin(GL_LINES); \
    { \
      glVertex2f( \
	(GLfloat)(offset_x + x2 + (tick_x_min * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 + (tick_x_min * sin_bank) + half_height) \
      ); \
      glVertex2f( \
	(GLfloat)(offset_x + x2 + ((5 + tick_x_max) * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 + ((5 + tick_x_max) * sin_bank) + half_height) \
      ); \
 \
      glVertex2f( \
	(GLfloat)(offset_x + x2 - (tick_x_min * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 - (tick_x_min * sin_bank) + half_height) \
      ); \
      glVertex2f( \
	(GLfloat)(offset_x + x2 - ((5 + tick_x_max) * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 - ((5 + tick_x_max) * sin_bank) + half_height) \
      ); \
    } \
    glEnd(); \
} \
else \
{ \
    glBegin(GL_LINES); \
    { \
      glVertex2f( \
	(GLfloat)(offset_x + x2 + (tick_x_min * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 + (tick_x_min * sin_bank) + half_height) \
      ); \
      glVertex2f( \
	(GLfloat)(offset_x + x2 + (tick_x_max * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 + (tick_x_max * sin_bank) + half_height) \
      ); \
 \
      glVertex2f( \
	(GLfloat)(offset_x + x2 - (tick_x_min * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 - (tick_x_min * sin_bank) + half_height) \
      ); \
      glVertex2f( \
	(GLfloat)(offset_x + x2 - (tick_x_max * cos_bank) + half_width), \
	(GLfloat)(offset_y + y2 - (tick_x_max * sin_bank) + half_height) \
      ); \
    } \
    glEnd(); \
 } \
 degi += tick_inc_degreesi; \
}


        degi = tick_start_pitch;

        /* Draw upper (pitch up) ticks. */
        for(y = tick_start_y; y < end_max; y += tick_inc_yi)   
        {
            if(degi > 90) 
                break;

            DO_DRAW_ATTITUDE_TICK
        }

        degi = tick_start_pitch - tick_inc_degreesi;

        /* Draw lower (pitch down) ticks. */
        for(y = tick_start_y - tick_inc_yi; y > end_min; y -= tick_inc_yi)
        {
            if(degi < -90)
                break;

            DO_DRAW_ATTITUDE_TICK
        }
#undef DO_DRAW_ATTITUDE_TICK


        /* Draw aircraft axis marker. */
        x = offset_x + half_width;
        y = offset_y + half_height;
	glBegin(GL_LINES);
	{
	    glVertex2f(
		(GLfloat)(x - 6),
		(GLfloat)y
	    );
	    glVertex2f(
		(GLfloat)(x + 6),
		(GLfloat)y
	    );
	}
	glEnd();

	/* Draw speed and direction label. */
	if(obj_aircraft_ptr != NULL)
	{
	    sar_obj_intercept_struct *intercept = NULL;
	    double intercept_dtheta = 0.0, intercept_sin_theta,
		intercept_cos_theta, intercept_dist = 0.0,
		vel_speed;


	    /* Get pointer to selected intercept. */
            i = obj_aircraft_ptr->cur_intercept;
            if((i >= 0) && (i < obj_aircraft_ptr->total_intercepts))
	    {
                intercept = obj_aircraft_ptr->intercept[i];
		if(intercept != NULL)
		{
		    double	icpt_dx = intercept->x - pos->x,
				icpt_dy = intercept->y - pos->y;

		    /* Calculate intercept delta theta in radians. */
                    intercept_dtheta = SFMDeltaRadians(
                        dir->heading,
                        SFMSanitizeRadians(
			    (0.5 * PI) - atan2(icpt_dy, icpt_dx)
                        )
                    );
		    /* Calculate intercept distance in meters. */
		    intercept_dist = hypot(icpt_dx, icpt_dy);
		}
	    }
	    intercept_sin_theta = sin(intercept_dtheta);
            intercept_cos_theta = cos(intercept_dtheta);

            /* Calculate 2D vel. */
            vel_speed = hypot(
                obj_aircraft_ptr->vel.x,
                obj_aircraft_ptr->vel.y
            );

	    /* Speed and distance to intercept label. */
	    if(intercept_dist > 0.0)
	    {
		double	v1 = SFMMPCToMPH(obj_aircraft_ptr->speed),
			v2 = SFMMetersToMiles(intercept_dist);

		x = offset_x + half_width - (width / 4) - (7 * 8);
		y = height - (offset_y + half_height - 4);
		if(v2 > 10.0)
		    sprintf(text, "%.1f MPH  %.0f", v1, v2);
		else
		    sprintf(text, "%.1f MPH  %.1f", v1, v2);
	    }
	    else
	    {
		double v1 = SFMMPCToMPH(obj_aircraft_ptr->speed);

		x = offset_x + half_width - (width / 4) - (7 * 5);
		y = height - (offset_y + half_height - 4);
		sprintf(text, "%.1f MPH", v1);
	    }
	    GWDrawString(display, x, y, text);

	    /* Draw velocity direction (object relative). */
            x = offset_x + half_width - (width / 4);
            y = offset_y + half_height + 12;
	    if(vel_speed > 0)
	    {
		glBegin(GL_LINES);
		{
		    glVertex2f(
			(GLfloat)(x + ((obj_aircraft_ptr->vel.x / vel_speed) * 12.0)),
			(GLfloat)(y + ((obj_aircraft_ptr->vel.y / vel_speed) * 12.0))
		    );
		    glVertex2f(
			(GLfloat)x,
			(GLfloat)y
		    );
		}
		glEnd();
	    }

	    /* Intercept direction. */
	    glBegin(GL_LINES);
	    {
		glVertex2f(
		    (GLfloat)(x + (intercept_sin_theta * 16.0)),
		    (GLfloat)(y + (intercept_cos_theta * 16.0))
		);
		glVertex2f(
		    (GLfloat)(x + (intercept_sin_theta * 10.0)),
		    (GLfloat)(y + (intercept_cos_theta * 10.0))
	        );
	    }
	    glEnd();

            glRasterPos2i(
		x - (vel_dir_w / 2) + 1,
		y - (vel_dir_h / 2)
	    );
            glBitmap(
                vel_dir_w, vel_dir_h, 
                0.0, 0.0,
                (GLfloat)vel_dir_w, 0.0,
                vel_dir_bm
            );
	}

	/* Draw altitude label. */
	if(obj_aircraft_ptr != NULL)
	{
	    double v1 = SFMMetersToFeet(
		(-obj_aircraft_ptr->center_to_ground_height) +
		obj_aircraft_ptr->cockpit_offset_pos.z
	    );

            x = offset_x + half_width + (width / 4) - (7 * 3);
            y = height - (offset_y + half_height + 32);

	    switch(obj_aircraft_ptr->engine_state)
	    {
	      case SAR_ENGINE_STATE_OFF:
                sprintf(text, "Eng(O): %.0f%%", obj_aircraft_ptr->throttle * 100);
                break;

	      case SAR_ENGINE_STATE_INIT:
		sprintf(text, "Eng(I): %.0f%%", obj_aircraft_ptr->throttle * 100);
		break;

	      default:
		sprintf(text, "Thr: %.0f%%", obj_aircraft_ptr->throttle * 100);
		break;
	    }
	    GWDrawString(display, x, y, text);


            y += 18;
            sprintf(text, "%.0f AGL", v1);
	    GWDrawString(display, x, y, text);

	    v1 = SFMMPCToMPH(obj_aircraft_ptr->vel.z);
	    y += 18;
	    if(v1 > 10.0)
	        sprintf(text, "+%.0f ROC", v1);
            else if(v1 > 0.0)
		sprintf(text, "+%.1f ROC", v1);
	    else if(v1 > -10.0)
		sprintf(text, "%.1f ROC", v1);
            else
                sprintf(text, "%.0f ROC", v1);
	    GWDrawString(display, x, y, text);
	}

	/* Draw heading label (just above attitude ticks). */
	if(obj_aircraft_ptr != NULL)
	{
	    double v1 = SFMRadiansToDegrees(dir->heading);

	    x = offset_x + half_width - 10;
	    y = height - (offset_y + half_height + 36);
	    sprintf(text, "%.0f", v1);
            GWDrawString(display, x, y, text);
	}

	/* Draw coordinates if viewing from map. */
	if(scene->camera_ref == SAR_CAMERA_REF_MAP)
	{
	    double dms_x, dms_y;


	    SFMMToDMS(
		camera_pos.x, camera_pos.y,
		scene->planet_radius,
		scene->dms_x_offset, scene->dms_y_offset,
		&dms_x, &dms_y
	    );

	    sprintf(
		text,
		"%s %s  Alt: %.0f",
		SFMLongitudeToString(dms_x),
                SFMLatitudeToString(dms_y),
		SFMMetersToFeet(camera_pos.z)
	    );
            GWDrawString(display, 5, 5, text);
	}

	return;
}


/*
 *	Draws on screen messages.
 */
void SARDrawMessages(
	gw_display_struct *display,
	sar_scene_struct *scene
)
{
	int i, len, offset_x, offset_y, width, height,
	    fw, fh, half_width, half_height, bg_width, bg_height;
	char *strptr;
	GWFont *font;

        /* Coordinates are handled in left hand rule xy plane,
         * the y axis is inverted at the call to the draw
         * function.
         */

	/* Display messages at all? */
	if(scene->message_display_untill <= 0)
	    return;

	/* Time to stop displaying messages? */
	if(scene->message_display_untill < cur_millitime)
	{
	    /* Reset message display untill time to 0 and delete
	     * all messages on the scene. This will not change
	     * the size of the message pointer array.
	     */
	    SARMessageClearAll(scene);
	    return;
	}

	/* Calculate dimensions. */
        width = display->width;
        height = display->height;

        half_width = width / 2;
        half_height = height / 2;

	font = option.message_font;
	GWGetFontSize(font, NULL, NULL, &fw, &fh);
	offset_x = 0;
        offset_y = (int)(height * 0.20);

	bg_width = width;
	bg_height = (scene->total_messages * fh);

	/* Draw darkened background. */
	StateGLEnable(&display->state_gl, GL_BLEND);
	StateGLBlendFunc(
	    &display->state_gl, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA
	);
	glBegin(GL_QUADS);
	{
	    glColor4f(0.0, 0.0, 0.0, 0.5);
	    glVertex2f(0, (GLfloat)offset_y);
	    glVertex2f((GLfloat)bg_width, (GLfloat)offset_y);
            glVertex2f((GLfloat)bg_width, (GLfloat)(offset_y + bg_height));
            glVertex2f(0, (GLfloat)(offset_y + bg_height));
	}
	glEnd();
	StateGLDisable(&display->state_gl, GL_BLEND);

	/* Begin drawing messages, first message is darkest color. */
	glColor4f(1.0, 1.0, 1.0, 1.0);
        GWSetFont(display, option.message_font);

	offset_y += fh;
	for(i = scene->total_messages - 1; i >= 0; i--)
	{ 
	    strptr = scene->message[i];
	    if(strptr == NULL)
	        break;

	    len = strlen(strptr);
	    offset_x = half_width - ((len * fw) / 2);

            GWDrawString(
		display,
		offset_x,
		height - offset_y,
		strptr
	    );

	    /* Lighten color. */
            glColor4f((GLfloat)0.8, (GLfloat)0.8, (GLfloat)0.8, (GLfloat)1.0);

	    offset_y += fh;
	}

	return;
}


/*
 *	Draws the light array.
 */
void SARDrawLights(
	gw_display_struct *display, sar_scene_struct *scene,
	sar_light_struct **ptr, int total
)
{
	int i;
	sar_light_struct *light_ptr;
	sar_color_struct *c;
	sar_position_struct *pos;


	if(total <= 0)
	    return;

	STATE_FOR_POINT_LIGHTS_BEGIN

	StateGLEnable(&display->state_gl, GL_BLEND);
	StateGLBlendFunc(
	    &display->state_gl, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA
	);
	StateGLEnable(&display->state_gl, GL_POINT_SMOOTH);

	for(i = 0; i < total; i++)
	{
	    light_ptr = ptr[i];
	    if(light_ptr == NULL)
		continue;

	    if(!(light_ptr->flags & SAR_LIGHT_FLAG_ON))
		continue;

	    /* Skip attenuation spot lights. */
	    if(light_ptr->flags & SAR_LIGHT_FLAG_ATTENUATE)
		continue;

	    if(light_ptr->flags & SAR_LIGHT_FLAG_STROBE)
	    {
		if(light_ptr->next_on > 0)
		{
		    /* Waiting to be turned on, implies it's off. */

		    if(light_ptr->next_on > cur_millitime)
		    {
			continue;
		    }
		    else
		    {
			light_ptr->next_on = 0;
			light_ptr->next_off = cur_millitime +
			    light_ptr->int_on;
		    }
		}
		else
		{
		    /* Waiting to be turned off, implies its on. */

		    if(light_ptr->next_off <= cur_millitime)
                    {
			light_ptr->next_on = cur_millitime +
			    light_ptr->int_off;
                        light_ptr->next_off = 0;

                        continue;
                    }
		}
	    }

	    pos = &light_ptr->pos;
	    c = &light_ptr->color;

            StateGLPointSize(
		&display->state_gl,
		(GLfloat)MAX(light_ptr->radius * 2.0, 1.0)
	    );
	    glBegin(GL_POINTS);
	    {
                glColor4f(
                    (GLfloat)c->r,
                    (GLfloat)c->g,
                    (GLfloat)c->b,
                    (GLfloat)c->a
                );
		glVertex3f((GLfloat)pos->x, (GLfloat)pos->z, (GLfloat)-pos->y);
	    }
	    glEnd();
	}

	StateGLDisable(&display->state_gl, GL_BLEND);

	STATE_FOR_POINT_LIGHTS_END


	return;
}


/*
 *	Draws cockpit (if available) on the object.  Translation and
 *	rotation is assumed to be at the object's center.
 */
void SARDrawObjectCockpit(
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr
)
{
        sar_object_aircraft_struct *obj_aircraft_ptr;


        switch(obj_ptr->type)
        {
          case SAR_OBJ_TYPE_AIRCRAFT:
            obj_aircraft_ptr = (sar_object_aircraft_struct *)
                obj_ptr->data;
            if(obj_aircraft_ptr == NULL)
                break;

	    SARVisualModelCallList(
		obj_aircraft_ptr->visual_model_cockpit
	    );
            break;

          default:
            break;
        }

	return;
}

/*
 *	Draws shadow (if available) on the object. Translation
 *	is already assumed to be at the object's center and heading
 *	(except pitch and bank) of object is already set.
 *
 *	Also draws rotor and prop wash if they are relivent for the
 *	given object's values.
 *
 *	Inputs assumed valid.
 */
void SARDrawObjectShadow(
	gw_display_struct *display, sar_scene_struct *scene,
	sar_object_struct *obj_ptr
)
{
        sar_object_aircraft_struct *obj_aircraft_ptr = NULL;
	double	throttle = 1.0,
		center_to_ground_height = 0.0;
	sar_visual_model_struct *vmodel = NULL;

	switch(obj_ptr->type)
	{
          case SAR_OBJ_TYPE_AIRCRAFT:
            obj_aircraft_ptr = (sar_object_aircraft_struct *)
                obj_ptr->data;
            if(obj_aircraft_ptr == NULL)
                break;

	    center_to_ground_height =
		obj_aircraft_ptr->center_to_ground_height;

	    throttle = obj_aircraft_ptr->throttle;

	    vmodel = obj_aircraft_ptr->visual_model_shadow;
	    break;

	  default:
	    break;
	}


	/* Draw prop wash? */
	if((obj_aircraft_ptr != NULL) &&
	   (center_to_ground_height >= -SAR_SHADOW_VISIBILITY_HEIGHT)
	)
	{
	    int i, tex_num, frame_num;
	    sar_obj_rotor_struct *rotor_ptr;
	    v3d_texture_ref_struct *t;
	    double r;
            double z = center_to_ground_height;

	    if(option.textured_objects)
	    {
		GLboolean depth_mask_flag;

                glPushMatrix();
                {
                    glTranslatef(0, (GLfloat)z, 0);

                    StateGLEnable(&display->state_gl, GL_BLEND);
                    StateGLDisable(&display->state_gl, GL_ALPHA_TEST);
                    depth_mask_flag = display->state_gl.depth_mask_flag;
                    StateGLDepthMask(&display->state_gl, GL_FALSE);  
                    StateGLEnable(&display->state_gl, GL_POLYGON_OFFSET_FILL);
		    StateGLBlendFunc(
			&display->state_gl, 
			GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA
		    );
                    StateGLPolygonOffset(
                        &display->state_gl,
			(GLfloat)option.gl_polygon_offset_factor, -1.0
                    );

		    for(i = 0; i < obj_aircraft_ptr->total_rotors; i++)
		    {
		        rotor_ptr = obj_aircraft_ptr->rotor[i];
		        if(rotor_ptr == NULL)
			    continue;

		        tex_num = rotor_ptr->rotor_wash_tex_num;
		        if(SARIsTextureAllocated(scene, tex_num))
			    t = scene->texture_ref[tex_num];
		        else
			    continue;

		        /* Get frame number. */
		        frame_num = t->total_frames *
			    rotor_ptr->rotor_wash_anim_pos /
			    ((sar_grad_anim_t)-1);
		        if(frame_num >= t->total_frames)
			    frame_num = t->total_frames - 1;
		        if(frame_num < 0)
			    frame_num = 0;

		        /* Select propwash texture. */
		        V3DTextureSelectFrame(t, frame_num);

			/* Set color, alpha based on distance to ground
			 * divided by shadow visibility height and
			 * multiplied by both global rotor wash visibility
			 * coeff and the throttle value.
			 */
		        glColor4f(
			    1.0, 1.0, 1.0,
			    (GLfloat)((1.0 + (double)(
				center_to_ground_height / SAR_SHADOW_VISIBILITY_HEIGHT)
			    ) * option.rotor_wash_vis_coeff * throttle)
			);
		        glShadeModel(GL_FLAT);

		        r = rotor_ptr->radius * 4;

                        glBegin(GL_QUADS);
                        {
                            glNormal3f(0, 1, 0);
                            glTexCoord2f(0, 1 - 0);
                            glVertex3f((GLfloat)-r, 0, (GLfloat)r);
                            glTexCoord2f(1, 1 - 0);
                            glVertex3f((GLfloat)r, 0, (GLfloat)r);
                            glTexCoord2f(1, 1 - 1);
                            glVertex3f((GLfloat)r, 0, (GLfloat)-r);
                            glTexCoord2f(0, 1 - 1);
                            glVertex3f((GLfloat)-r, 0, (GLfloat)-r);
                        }
                        glEnd();

		    }

                    StateGLDisable(&display->state_gl, GL_POLYGON_OFFSET_FILL);
                    StateGLDisable(&display->state_gl, GL_BLEND);
                    StateGLDepthMask(&display->state_gl, depth_mask_flag);
		    StateGLEnable(&display->state_gl, GL_ALPHA_TEST);
                }
                glPopMatrix();
	    }
	}

        /* Draw shadow? */
        if((vmodel != NULL) &&
           (center_to_ground_height >= -SAR_SHADOW_VISIBILITY_HEIGHT)
        )
        {
	    GLboolean depth_mask_flag;
            double z = center_to_ground_height;

            StateGLEnable(&display->state_gl, GL_BLEND);
            StateGLDisable(&display->state_gl, GL_ALPHA_TEST);
            depth_mask_flag = display->state_gl.depth_mask_flag;
            StateGLDepthMask(&display->state_gl, GL_FALSE);
            StateGLEnable(&display->state_gl, GL_POLYGON_OFFSET_FILL);
            StateGLBlendFunc(
                &display->state_gl, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA
            );
            StateGLPolygonOffset(
                &display->state_gl, 
		(GLfloat)option.gl_polygon_offset_factor, -1.0
            );

            glPushMatrix();
            {
                glTranslatef(0, (GLfloat)z, 0);
                SARVisualModelCallList(vmodel);
            }
            glPopMatrix();

            StateGLDisable(&display->state_gl, GL_POLYGON_OFFSET_FILL);
            StateGLDisable(&display->state_gl, GL_BLEND);
	    StateGLDepthMask(&display->state_gl, depth_mask_flag);
            StateGLEnable(&display->state_gl, GL_ALPHA_TEST);
        }

	return;
}


/*
 *	Draws scene foundations (ground base).
 *
 *	Inputs assumed valid.
 */
void SARDrawSceneFoundations(
	gw_display_struct *display, sar_scene_struct *scene,
	double aspect
)
{
	int in_map_view;
	int mod_offset_x, mod_offset_y;
	sar_position_struct *pos;
	double fovx, fovz;	/* Field of view x and z, in radians. */
        sar_cloud_layer_struct *cloud_layer_ptr = NULL;
	sar_visual_model_struct *vmodel = NULL;


        pos = &camera_pos;

	/* Get field of views in radians. */
	fovz = scene->camera_fovz;
        if(fovz <= 0)
            return;
        fovx = fovz * aspect;

	/* In map view? */
	in_map_view = ((scene->camera_ref == SAR_CAMERA_REF_MAP) ?
            1 : 0
	);

        /* Get pointer to first (lowest) cloud layer. */
	cloud_layer_ptr = lowest_cloud_layer_ptr;

	/* Cloud layer valid and not in map view? */
        if((cloud_layer_ptr != NULL) && !in_map_view)
        {
            /* If we are at or above the first cloud layer,
             * then do not draw foundations since the would not
	     * be visable.
             */
            if(pos->z >= cloud_layer_ptr->z)
                return;
        }

	/* Calculate module offset. */
        mod_offset_x = (int)pos->x %
            (int)scene->base.tile_width;
        mod_offset_y = (int)pos->y %
            (int)scene->base.tile_height;


	glPushMatrix();
	{
	    /* Translate to new tiled position. */
 	    glTranslatef(
                (GLfloat)((int)pos->x - mod_offset_x),
                0,
                (GLfloat)-((int)pos->y - mod_offset_y)
            );

	    /* Draw simple (solid color) base visual model. */
	    vmodel = scene->base.visual_model_simple;

	    /* Is visual model valid and not in map view? */
	    if((vmodel != NULL) && !in_map_view)
	    {
		SAR_DRAW_TEXTURE_1D_OFF
		SAR_DRAW_TEXTURE_2D_OFF
                glColor4f(
                    (GLfloat)(scene->base.color.r * scene->light_color.r),
                    (GLfloat)(scene->base.color.g * scene->light_color.g),
                    (GLfloat)(scene->base.color.b * scene->light_color.b),
                    1.0
                );
		SARVisualModelCallList(vmodel);
	    }

	    /* Draw close (textured) tiles. */
	    vmodel = scene->base.visual_model_close;
	    if((vmodel != NULL) && option.textured_ground)
	    {
		SAR_DRAW_TEXTURE_2D_ON
                glColor4f(
                    (GLfloat)scene->light_color.r,
                    (GLfloat)scene->light_color.g,
                    (GLfloat)scene->light_color.b,
                    1.0
                );
		SARVisualModelCallList(vmodel);
	    }
	}
	glPopMatrix();


	return;
}

/*
 *	Draws cloud layer.
 *
 *	Inputs assumed valid.
 */
void SARDrawCloudLayer(
	sar_scene_struct *scene,
        sar_cloud_layer_struct *cloud_layer_ptr,
	Boolean flip
)
{
        int mod_offset_x, mod_offset_y;
        sar_position_struct *pos;


	if(cloud_layer_ptr->visual_model == NULL)
	    return;

        pos = &camera_pos;

	mod_offset_x = (int)pos->x %
	    (int)cloud_layer_ptr->tile_width;
	mod_offset_y = (int)pos->y %
	    (int)cloud_layer_ptr->tile_height;

        glPushMatrix();
        {
            glTranslatef(
                (GLfloat)((int)pos->x - mod_offset_x),
                (GLfloat)cloud_layer_ptr->z,
                (GLfloat)-((int)pos->y - mod_offset_y)
            );
	    if(flip)
		glRotatef(
		    (GLfloat)180, 0.0, 0.0, 1.0
		);

	    glColor4f(
		(GLfloat)scene->light_color.r,
                (GLfloat)scene->light_color.g,
                (GLfloat)scene->light_color.b,
		1.0
	    );
	    SARVisualModelCallList(cloud_layer_ptr->visual_model);
        }
        glPopMatrix();

	return;
}

/*
 *	Draws horizon quads, translation needs to be at scene
 *	origin.
 *
 *	Inputs assumed valid and depth test and lighting turned off.
 */
void SARDrawSceneHorizon(
        gw_display_struct *display, sar_scene_struct *scene
)
{
	GLboolean depth_mask_flag;
	int	i, n,
		afternoon = 0,
		total_horizon_textures = 7,
		total_horizon_panels = 12;
	double h_min, h_max, r;
	sar_position_struct *pos;
	sar_color_struct *c;
	sar_scene_horizon_struct *horizon_ptr = &scene->horizon;
	sar_cloud_layer_struct *cloud_layer_ptr = NULL;


        /* Get camera position. */
        pos = &camera_pos;
	c = &scene->light_color;

	/* Match cloud layer just above camera. */
	for(i = 0; i < scene->total_cloud_layers; i++)
	{
	    cloud_layer_ptr = scene->cloud_layer[i];
	    if(cloud_layer_ptr == NULL)
		continue;

	    if(cloud_layer_ptr->z > pos->z)
		break;
	    else
		cloud_layer_ptr = NULL;
	}


	/* Check if horizon has all 7 textures allocated. */
	if(horizon_ptr->total_textures < total_horizon_textures)
	{
	    /* Not enough textures allocated. */
	    return;
	}

	/* Calculate horizon radius r, this is the distance from camera
	 * to each horizon panel. Note that we make this distance a bit
	 * shorter than the maximum visibility to ensure that the horizon
	 * panels are drawn closer than the camera's far clip.
	 */
	r = visibility_max * 0.85;

	/* Calculate height minimum, always 0.0. */
	h_min = 0.0;

	/* Calculate height maximum, dependant of cloud layer existance.
	 * If a cloud layer exists then it will be 1.05 times the cloud
	 * layer's altitude (we use 1.05 to ensure that it goes up into
	 * the clouds just a bit). Otherwise it will be calculated by
	 * s = r * theta where theta is (0.25 * PI).
	 */
	if(cloud_layer_ptr != NULL)
	    h_max = cloud_layer_ptr->z * 1.05;	/* Add a bit of margin. */
	else
	    h_max = MAX(r * (0.25 * PI), pos->z);

	/* Make note if it is afternoon or later. */
	if(scene->tod > (12 * 3600))
	    afternoon = 1;

	SAR_DRAW_TEXTURE_1D_ON
	SAR_DRAW_TEXTURE_2D_OFF

        glShadeModel(GL_FLAT);

	/* No need to write depth when drawing horizon panels, this
	 * will allow some objects to be drawn farther away from
	 * the horizon panels.
	 */
	depth_mask_flag = display->state_gl.depth_mask_flag;
	StateGLDepthMask(&display->state_gl, GL_FALSE);

        glPushMatrix();
        {
	    /* Angle in radians of the current angle from camera to the
	     * the horizon panel's `previous' point.
	     */
	    double theta;

	    /* Angle in radians of normal. */
	    double n_theta;

	    /* Deltas of each panel point along xy plane where index 0
	     * is the `previous point' going clockwise from 9 o'clock.
	     */
	    double dx[2], dy[2];


	    /* Translate to camera position but at z = 0.0. */
            glTranslatef(
		(GLfloat)pos->x, (GLfloat)0.0, (GLfloat)-pos->y
	    );

	    /* Set color to be the scene's light color so that it
	     * controls the brightness of the horizon.
	     */
            glColor4f(
                (GLfloat)c->r,
                (GLfloat)c->g,
                (GLfloat)c->b,
                (GLfloat)1.0
            );

	    /* Initialize theta. */
	    theta = SFMSanitizeRadians(
		(1.5 * PI) -
		(((2.0 * PI) / total_horizon_panels) / 2)
	    );

	    /* Initialize `previous point' for the panel delta offset. */
	    dx[0] = r * sin(theta);
	    dy[0] = r * cos(theta);

	    /* Begin drawing each horizon panel. */
	    for(i = 0; i < total_horizon_panels; i++)
	    {
		/* Increment theta to `next' angle first. */
		theta = SFMSanitizeRadians(
		    theta +
		    ((2.0 * PI) / total_horizon_panels)
		);

		/* Calculate `next' panel point. */
		dx[1] = r * sin(theta);
		dy[1] = r * cos(theta);


                /* Calculate horizon texture index. */
		if(i >= total_horizon_textures)
		    n = total_horizon_textures -
			i + total_horizon_textures - 2;
		else
		    n = i;
		/* Clip. */
		if((n >= 0) && (n < total_horizon_textures))
		    V3DTextureSelect(horizon_ptr->texture[n]);

		/* Calculate normal theta. */
		n_theta = SFMSanitizeRadians(
                    theta -
                    (((2.0 * PI) / total_horizon_panels) / 2) +
		    (1.0 * PI)
                );

		/* Begin drawing this horizon panel as a GL_QUADS. */
		glBegin(GL_QUADS);
		glNormal3f(
		    (GLfloat)sin(n_theta),
		    0.0,
		    -(GLfloat)cos(n_theta)
		);		    

                glTexCoord2f(1, 0);
                glVertex3f(
		    (GLfloat)dx[0], (GLfloat)h_min, (GLfloat)-dy[0]
		);
                glTexCoord2f(1, 0);
                glVertex3f(
		    (GLfloat)dx[1], (GLfloat)h_min, (GLfloat)-dy[1]
		);
                glTexCoord2f(0, 0);
                glVertex3f(
		    (GLfloat)dx[1], (GLfloat)h_max, (GLfloat)-dy[1]
		);
                glTexCoord2f(0, 0);
                glVertex3f(
		    (GLfloat)dx[0], (GLfloat)h_max, (GLfloat)-dy[0]
		);
		glEnd();
		/* Done drawing this horizon panel. */

		/* Set next panel point to previous panel point. */
		dx[0] = dx[1];
		dy[0] = dy[1];
	    }	/* Begin drawing each horizon panel. */
        }
        glPopMatrix();

	/* Restore depth writing to what it was. */
        StateGLDepthMask(&display->state_gl, depth_mask_flag);

	return;
}


/*
 *	Redraws scene.
 */
void SARDraw(sar_core_struct *core_ptr)
{
        int i, n, *total;
	GLfloat view_aspect;
	Boolean need_clear_color = True;
	gw_display_struct *display;
	sar_scene_struct *scene;
        sar_object_struct ***ptr, *obj_ptr;
	gctl_struct *gc;
	sar_direction_struct *dir;
	sar_position_struct *pos;
	sar_object_aircraft_struct *obj_aircraft_ptr;
	sar_object_helipad_struct *obj_helipad_ptr;
	sar_object_runway_struct *obj_runway_ptr;
	sar_object_human_struct *obj_human_ptr;
	sar_obj_rotor_struct *rotor_ptr;
	sar_obj_landing_gear_struct *lgear_ptr;
	sar_obj_hoist_struct *hoist_ptr;
	sar_cloud_layer_struct *cloud_layer_ptr;
	double distance, distance3d;
	GLfloat light_val[4];


	/* Get initial pointers to resources. */
	if(core_ptr == NULL)
	     return;

	display = core_ptr->display;
	if(display == NULL)
	    return;

	scene = core_ptr->scene;
	ptr = &core_ptr->object;
	total = &core_ptr->total_objects;
	gc = core_ptr->gctl;

	if(scene == NULL)
	    return;

        /* Draw map in a different function. */
        if(scene->camera_ref == SAR_CAMERA_REF_MAP)
        {
	    SARDrawMap(core_ptr, False);
	    return;
        }

	/* Calculate maximum visibility. */
	visibility_max = SFMMilesToMeters(3 + (option.visibility_max * 3));

	/* Reset translations. */
#define DRAW_SET_PERSPECTIVE_FAR	\
glMatrixMode(GL_PROJECTION); \
glLoadIdentity(); \
gluPerspective( \
 (GLfloat)SFMRadiansToDegrees(scene->camera_fovz), /* Field of view. */ \
 view_aspect, \
 (GLfloat)0.2, \
 (GLfloat)visibility_max \
); \
glMatrixMode(GL_MODELVIEW);

#define DRAW_SET_PERSPECTIVE_CLOSE	\
glMatrixMode(GL_PROJECTION); \
glLoadIdentity(); \
gluPerspective( \
 SFMRadiansToDegrees(scene->camera_fovz), /* Field of view. */ \
 view_aspect, \
 0.1, \
 1000 \
); \
glMatrixMode(GL_MODELVIEW);

	SARReshapeCB(display, -1, -1, -1, -1);
	if(display->height > 0)
	    view_aspect = (GLfloat)display->width / (GLfloat)display->height;
	else
	    view_aspect = (800 / 600);
	DRAW_SET_PERSPECTIVE_FAR
        glLoadIdentity();


	/* Set camera location, this will set camera position variables
	 * for this source file.
	 */
	SARSetCamera(scene, ptr, total);

	/* Reset other variables. */
	player_wheel_brakes = 0;
	player_air_brakes = 0;

	/* Get pointer to lowest (first) cloud layer. */
	i = scene->total_cloud_layers;
	if(i > 0)
	{
	    lowest_cloud_layer_ptr = scene->cloud_layer[0];
	    highest_cloud_layer_ptr = scene->cloud_layer[i - 1];
	}
	else
	{
	    lowest_cloud_layer_ptr = NULL;
	    highest_cloud_layer_ptr = NULL;
	}

        /* Atmosphere enabled? */
        if(option.atmosphere)
        {
	    /* Atmosphere is enabled, set up atmosphere and sky color. */
            GLfloat fog_color[4];
            double g;

	    /* Get average of light color as gamma. */         
            g = (scene->light_color.r + scene->light_color.g +
                scene->light_color.b) / 3;

	    /* Reset sky color and take atmosphere into account. */
            if(highest_cloud_layer_ptr != NULL)
            {
                if(camera_pos.z < highest_cloud_layer_ptr->z)
                {
                    /* Below the highest cloud layer. */

		    /* Since below the highest cloud layer and
		     * if textured clouds are enabled, then do not
		     * clear the color buffer since drawing the cloud
		     * layers, horizon, and ground base will ensure that
		     * the frame buffer is `effectivly cleared'.
		     */
		    if(option.textured_clouds)
			need_clear_color = False;


                    sky_color[0] = (GLfloat)(
			scene->sky_nominal_color.r * pow(g, 0.5)
		    );
                    sky_color[1] = (GLfloat)(
			scene->sky_nominal_color.g * g
		    );
                    sky_color[2] = (GLfloat)(
			scene->sky_nominal_color.b * g
		    );
                }
                else
                {
                    /* Above the highest cloud layer. */
                    sky_color[0] = (GLfloat)(
			scene->sky_nominal_color.r * 0.15 * g
		    );
                    sky_color[1] = (GLfloat)(
			scene->sky_nominal_color.g * 0.15 * g
		    );
                    sky_color[2] = (GLfloat)(
			scene->sky_nominal_color.b * 0.15 * g
		    );
                }
            }
	    else
	    {
                sky_color[0] = (GLfloat)(
                    scene->sky_nominal_color.r * pow(g, 0.5)
                );
                sky_color[1] = (GLfloat)(
                    scene->sky_nominal_color.g * g
                );
                sky_color[2] = (GLfloat)(
                    scene->sky_nominal_color.b * g
                );
            }

	    /* Set atmosphere color. */
            atmosphere_color[0] = (GLfloat)MIN(sky_color[0] * 1.1, 1.0);
            atmosphere_color[1] = (GLfloat)MIN(sky_color[1] * 1.1, 1.0);
            atmosphere_color[2] = (GLfloat)MIN(sky_color[2] * 1.1, 1.0);

	    /* Set fog color. */
            fog_color[0] = atmosphere_color[0];
            fog_color[1] = atmosphere_color[1];
            fog_color[2] = atmosphere_color[2];
            fog_color[3] = 1.0;

	    StateGLEnable(&display->state_gl, GL_FOG);
            glFogi(GL_FOG_MODE, GL_EXP);
            glFogfv(GL_FOG_COLOR, fog_color);
            glFogf(GL_FOG_DENSITY,
		(GLfloat)(
		    (1.0 / visibility_max) *
		    (8.0 * scene->atmosphere_density_coeff)
		)
	    );
	    glHint(GL_FOG_HINT, GL_DONT_CARE);
            glFogf(GL_FOG_START,
                (GLfloat)(
		    visibility_max *
		    (1.0 - scene->atmosphere_dist_coeff)
                )
            );
            glFogf(GL_FOG_END, (GLfloat)visibility_max);

	    glClearColor(
                (GLclampf)atmosphere_color[0],
                (GLclampf)atmosphere_color[1],
                (GLclampf)atmosphere_color[2],
                (GLclampf)0.0
            );
        }
	else
	{
	    /* Atmosphere is disabled, just set sky color. */

            /* Reset sky color and take primary light color
	     * into account.
	     */
            if(highest_cloud_layer_ptr != NULL)
            {
                if(camera_pos.z < highest_cloud_layer_ptr->z)
                {
                    /* Below the highest cloud layer. */

                    /* Since below the highest cloud layer and
                     * if textured clouds are present, then do not
                     * clear the color buffer.
                     */
                    if(option.textured_clouds)
                        need_clear_color = False;

                    sky_color[0] = (GLfloat)(
                        scene->sky_nominal_color.r * scene->light_color.r
                    );
                    sky_color[1] = (GLfloat)(
                        scene->sky_nominal_color.g * scene->light_color.g
                    );
                    sky_color[2] = (GLfloat)(
                        scene->sky_nominal_color.b * scene->light_color.b
                    );
                }
                else
                {
                    /* Above the highest cloud layer. */
                    sky_color[0] = (GLfloat)(
                        scene->sky_nominal_color.r * 0.15 * scene->light_color.r
                    );
                    sky_color[1] = (GLfloat)(
                        scene->sky_nominal_color.g * 0.15 * scene->light_color.g
                    );
                    sky_color[2] = (GLfloat)(
                        scene->sky_nominal_color.b * 0.15 * scene->light_color.b
                    );
                }
            }
            else
            {
                sky_color[0] = (GLfloat)(
                    scene->sky_nominal_color.r * scene->light_color.r
                );
                sky_color[1] = (GLfloat)(
                    scene->sky_nominal_color.g * scene->light_color.g
                );
                sky_color[2] = (GLfloat)(
                    scene->sky_nominal_color.b * scene->light_color.b
                );
            }

            glClearColor(
                (GLclampf)sky_color[0],
                (GLclampf)sky_color[1],
                (GLclampf)sky_color[2],
                (GLclampf)0.0
            );
 	}

        /* Clear GL buffer. */
	glClearDepth((GLclampd)1.0);
	/* Clear color set above in fog check. */
        glClear(
            ((need_clear_color) ? GL_COLOR_BUFFER_BIT : 0) |
            GL_DEPTH_BUFFER_BIT
        );


	/* Disable lighting and depth test. */
	StateGLDisable(&display->state_gl, GL_LIGHTING);
	StateGLDisable(&display->state_gl, GL_LIGHT0);
	/* GL_LIGHT1 is turned on or off later. */

	SAR_DRAW_DEPTH_TEST_OFF

	SAR_DRAW_TEXTURE_1D_OFF

	SAR_DRAW_TEXTURE_2D_OFF


	/* Draw ground base. */
        SARDrawSceneFoundations(display, scene, view_aspect);

        /* Draw cloud layers and textured horizon? */
        if(option.textured_clouds)
        {
	    /* Yes, draw cloud layers and horizon. */

	    /* Horizon. */
            if(highest_cloud_layer_ptr != NULL)
            {
                if(camera_pos.z < highest_cloud_layer_ptr->z)
                    SARDrawSceneHorizon(display, scene);
            }
            else 
            {   
                SARDrawSceneHorizon(display, scene);
            }

	    SAR_DRAW_TEXTURE_1D_OFF
	    SAR_DRAW_TEXTURE_2D_ON

	    /* Note, cloud layers should be allocated lowest to
	     * highest.
	     */

	    /* Go through cloud layers above camera. */
            for(i = 0; i < scene->total_cloud_layers; i++)
            {
                cloud_layer_ptr = scene->cloud_layer[i];
                if(cloud_layer_ptr == NULL)
                    continue;

                /* Above camera? */
                if(cloud_layer_ptr->z > camera_pos.z)
                {
		    /* Draw this cloud later and that's it. */
                    SARDrawCloudLayer(scene, cloud_layer_ptr, True);
                    break;
                }
            }
	    /* Go through cloud layers below camera. */
            for(i = scene->total_cloud_layers - 1; i >= 0; i--)
            {
		cloud_layer_ptr = scene->cloud_layer[i];
                if(cloud_layer_ptr == NULL)
                    continue;

		/* Below camera? */
		if(cloud_layer_ptr->z < camera_pos.z)
		{
		    /* Draw this cloud later and that's it. */
                    SARDrawCloudLayer(scene, cloud_layer_ptr, False);
		    break;
		}
            }
        }
	else
	{
	    /* Do not draw lower cloud layers or horizon, however we
	     * still must draw the highest cloud layer if it
	     * is defined and the camera is above it.
	     */
	    if(highest_cloud_layer_ptr != NULL)
            {
                if(camera_pos.z > highest_cloud_layer_ptr->z)
		{
		    SAR_DRAW_TEXTURE_2D_ON
                    SARDrawCloudLayer(scene, highest_cloud_layer_ptr, False);
		}
            }

	}

	/* Enable lighting and turn on global GL_LIGHT0. */
        StateGLEnable(&display->state_gl, GL_LIGHTING);
	StateGLEnable(&display->state_gl, GL_LIGHT0);
	/* Leave GL_LIGHT1 (search light) alone for now, it will be set
	 * later.
	 */

	/* Enable color material for front face only with lighting
	 * of ambience and diffusion.
	 */
	StateGLEnable(&display->state_gl, GL_COLOR_MATERIAL);
	StateGLColorMaterial(
	    &display->state_gl, GL_FRONT, GL_AMBIENT_AND_DIFFUSE
	);

	/* Set light 0 (global) position and values. */
	light_val[0] = (GLfloat)scene->light_pos.x;
	light_val[1] = (GLfloat)scene->light_pos.z;
	light_val[2] = (GLfloat)-scene->light_pos.y;
	light_val[3] = 1.0;
        glLightfv(GL_LIGHT0, GL_POSITION, &(light_val[0]));

	light_val[0] = (GLfloat)(0.40 * scene->light_color.r);
        light_val[1] = (GLfloat)(0.40 * scene->light_color.g);
        light_val[2] = (GLfloat)(0.40 * scene->light_color.b);
        light_val[3] = 1.0;
	glLightfv(GL_LIGHT0, GL_AMBIENT, &(light_val[0]));

	/* Set diffuse as the scene's lighting, increases it just a
	 * bit to make it last brighter around noon.
	 */
        light_val[0] = (GLfloat)CLIP(scene->light_color.r + 0.1, 0.0, 1.0);
        light_val[1] = (GLfloat)CLIP(scene->light_color.g + 0.1, 0.0, 1.0);
        light_val[2] = (GLfloat)CLIP(scene->light_color.b + 0.1, 0.0, 1.0);
        light_val[3] = 1.0;
        glLightfv(GL_LIGHT0, GL_DIFFUSE, &(light_val[0]));

        light_val[0] = 1.0;
        glLightfv(GL_LIGHT0, GL_CONSTANT_ATTENUATION, &(light_val[0]));

        light_val[0] = 0.0;
        glLightfv(GL_LIGHT0, GL_LINEAR_ATTENUATION, &(light_val[0]));

        light_val[0] = 0.0;
        glLightfv(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, &(light_val[0]));


	/* Turn on/off and set up texturing. */
        if(option.textured_objects)
	{
	    StateGLEnable(&display->state_gl, GL_ALPHA_TEST);
	    StateGLAlphaFunc(&display->state_gl, GL_GREATER, 0.5);

	    SAR_DRAW_TEXTURE_2D_ON
	}
        else
	{
	    SAR_DRAW_TEXTURE_2D_OFF
	}
	/* Don't use 1D texture. */
	SAR_DRAW_TEXTURE_1D_OFF


        /* Begin drawing each object. */
        for(i = 0; i < core_ptr->total_objects; i++)
        {
            obj_ptr = core_ptr->object[i];
            if(obj_ptr == NULL)
                continue;

	    /* Get pointer to position of object. */
            pos = &obj_ptr->pos;

            /* Camera above and object below lowest cloud layer? */
            if(lowest_cloud_layer_ptr != NULL)
            {
                if((camera_pos.z > lowest_cloud_layer_ptr->z) &&
                   (pos->z < lowest_cloud_layer_ptr->z)
                )
                    continue;
            }

            /* Check if object is to have depth testing or not. */
            if(obj_ptr->flags & SAR_OBJ_FLAG_NO_DEPTH_TEST)
            {
		SAR_DRAW_DEPTH_TEST_OFF
            }
            else
            {
		SAR_DRAW_DEPTH_TEST_ON
            }

	    /* Handle by object type. */
	    switch(*(int *)(obj_ptr))
	    {
              /* ************************************************** */
              case SAR_OBJ_TYPE_AIRCRAFT:
                obj_aircraft_ptr = obj_ptr->data;
		if(obj_aircraft_ptr == NULL)
		    break;

		/* Get position and check if in range with camera. */
		distance = hypot(pos->x - camera_pos.x, pos->y - camera_pos.y);
		if(distance > obj_ptr->range)
		    break;

		/* Get direction. */
		dir = &obj_ptr->dir;

		/* Get pointer to hoist. */
		hoist_ptr = &obj_aircraft_ptr->hoist;

/* Draws hoist, hoist_ptr must be valid. It will be checked if rope is
 * out which implies basket is out. Will modify pos and dir pointer 
 * values.
 */
#define DO_DRAW_HOIST_AS_NEEDED \
if(hoist_ptr->rope_cur > 0.0) \
{ \
 sar_position_struct *hoist_pos = &hoist_ptr->pos; \
 sar_direction_struct *hoist_dir = &hoist_ptr->dir; \
 glPushMatrix(); \
 { \
  /* Translate and rotate heading. */ \
  glTranslatef( \
   (GLfloat)hoist_pos->x, (GLfloat)hoist_pos->z, (GLfloat)-hoist_pos->y \
  ); \
  glRotatef((GLfloat)-SFMRadiansToDegrees(hoist_dir->heading), 0.0, 1.0, 0.0); \
 \
  SARDrawHoist(scene, obj_ptr, hoist_ptr, pos->z - hoist_pos->z); \
 } \
 glPopMatrix(); \
}

		glPushMatrix();
		{
		    /* Translate and rotate heading. */
		    glTranslatef((GLfloat)pos->x, (GLfloat)pos->z, (GLfloat)-pos->y);
		    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->heading), 0.0, 1.0, 0.0);

		    /* Draw shadow. */
                    SARDrawObjectShadow(display, scene, obj_ptr);

		    /* Rotate pitch and bank. */
		    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->pitch), 1.0, 0.0, 0.0);
		    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->bank), 0.0, 0.0, 1.0);

		    /* Calculate 3D distance. */
		    distance3d = hypot(distance, pos->z - camera_pos.z);

                    /* Check if this is the player object. */
                    if(obj_ptr == scene->player_obj_ptr)
                    {
			if(gc != NULL)
			{
			    player_wheel_brakes = gc->wheel_brakes_state;
			    if((player_wheel_brakes == 0) &&
			       (gc->wheel_brakes_coeff > 0.0)
			    )
				player_wheel_brakes = 1;

			    player_air_brakes = gc->air_brakes_state;
                            if((player_air_brakes == 0) &&
                               (gc->air_brakes_coeff > 0.0)
                            )
                                player_air_brakes = 1;
			}

                        /* Adjust engine volume. */
                        SARSoundAdjustEngineVolume(
                            core_ptr->recorder,
                            obj_aircraft_ptr->engine_inside_sndplay,
                            obj_aircraft_ptr->engine_outside_sndplay,
			    (scene->camera_ref == SAR_CAMERA_REF_COCKPIT),
                            obj_aircraft_ptr->engine_state,
                            obj_aircraft_ptr->throttle,
                            distance3d
                        );

			/* Adjust spot light. */
			SARDrawSetSpotLight(display, scene, obj_ptr);

                        if(scene->camera_ref == SAR_CAMERA_REF_COCKPIT)
			{
			    DRAW_SET_PERSPECTIVE_CLOSE

			    SARDrawObjectCockpit(scene, obj_ptr);

			    DRAW_SET_PERSPECTIVE_FAR

			    glPopMatrix();
			    break;
			}
		    }
		    else
		    {
                        /* Adjust engine volume. */
                        SARSoundAdjustEngineVolume(
                            core_ptr->recorder,
                            obj_aircraft_ptr->engine_inside_sndplay,
                            obj_aircraft_ptr->engine_outside_sndplay,
                            0,
                            obj_aircraft_ptr->engine_state,
                            obj_aircraft_ptr->throttle,
                            distance3d
                        );
		    }

		    /* Draw standard day time model? */
		    if((obj_ptr->flags & SAR_OBJ_FLAG_HIDE_DAY_MODEL) ?
			(scene->tod_code == SARSceneTODCodeDay) : 1
		    )
		    {
			SARVisualModelCallList(obj_ptr->visual_model);
		    }
                    /* Draw standard dawn time model? */
                    if((obj_ptr->flags & SAR_OBJ_FLAG_HIDE_DAWN_MODEL) ?
                        (scene->tod_code == SARSceneTODCodeDawn) : 1
                    )
                    {
                        /* Disable GL_LIGHTING when drawing dawn time model. */
                        StateGLDisable(&display->state_gl, GL_LIGHTING);
                        SARVisualModelCallList(obj_ptr->visual_model_dawn);
                        StateGLEnable(&display->state_gl, GL_LIGHTING);
                    }
                    /* Draw standard dusk time model? */
                    if((obj_ptr->flags & SAR_OBJ_FLAG_HIDE_DUSK_MODEL) ?
                        (scene->tod_code == SARSceneTODCodeDusk) : 1
                    )
		    {
			/* Disable GL_LIGHTING when drawing dusk time model. */
                        StateGLDisable(&display->state_gl, GL_LIGHTING);
                        SARVisualModelCallList(obj_ptr->visual_model_dusk);
			StateGLEnable(&display->state_gl, GL_LIGHTING);
		    }
                    /* Draw standard night time model? */
                    if((obj_ptr->flags & SAR_OBJ_FLAG_HIDE_NIGHT_MODEL) ?
                        (scene->tod_code == SARSceneTODCodeNight) : 1
                    )
		    {
			/* Disable GL_LIGHTING when drawing night time model. */
			StateGLDisable(&display->state_gl, GL_LIGHTING);
                        SARVisualModelCallList(obj_ptr->visual_model_night);
			StateGLEnable(&display->state_gl, GL_LIGHTING);
		    }

                    /* Draw rotors/propellars. */
                    for(n = 0; n < obj_aircraft_ptr->total_rotors; n++)
                    {
                        rotor_ptr = obj_aircraft_ptr->rotor[n];
                        if(rotor_ptr == NULL)
                            continue;

                        SARDrawRotor(
                            scene, obj_ptr, rotor_ptr
                        );
                    }

		    /* Draw landing gears. */
		    for(n = 0; n < obj_aircraft_ptr->total_landing_gears; n++)
		    {
			lgear_ptr = obj_aircraft_ptr->landing_gear[n];
			if(lgear_ptr == NULL)
			    continue;

			SARDrawLandingGear(
			    scene, obj_ptr, lgear_ptr
			);
		    }

		    /* Draw rescue door. */
		    SARDrawDoor(
			scene, obj_ptr,
			&obj_aircraft_ptr->rescue_door
		    );

		    POST_CALLLIST_RESET_STATES

                    /* Draw lights. */
                    SARDrawLights(  
                        display, scene,
                        obj_ptr->light, obj_ptr->total_lights  
                    );
		}
                glPopMatrix();

		/* Draw hoist as needed. */
		DO_DRAW_HOIST_AS_NEEDED

		break;

              /* ***************************************************** */
              case SAR_OBJ_TYPE_GROUND:
                /* Get position and check if in range with camera. */
                distance = hypot(pos->x - camera_pos.x, pos->y - camera_pos.y);
                if(distance > obj_ptr->range)
                    break;

                /* Get direction. */
                dir = &obj_ptr->dir;

                glPushMatrix();
                {
                    /* Translate and rotate. */
                    glTranslatef((GLfloat)pos->x, (GLfloat)pos->z, (GLfloat)-pos->y);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->heading), 0.0, 1.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->pitch), 1.0, 0.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->bank), 0.0, 0.0, 1.0);

                    /* Draw standard day time model? */
                    if((obj_ptr->flags & SAR_OBJ_FLAG_HIDE_DAY_MODEL) ?
                        (scene->tod_code == SARSceneTODCodeDay) : 1
                    )
                    {
                        SARVisualModelCallList(obj_ptr->visual_model);
                    }
                    /* Draw standard dawn time model? */
                    if((obj_ptr->flags & SAR_OBJ_FLAG_HIDE_DAWN_MODEL) ?
                        (scene->tod_code == SARSceneTODCodeDawn) : 1
                    )
                    {
                        /* Disable GL_LIGHTING when drawing dawn time model. */
                        StateGLDisable(&display->state_gl, GL_LIGHTING);
                        SARVisualModelCallList(obj_ptr->visual_model_dawn);
                        StateGLEnable(&display->state_gl, GL_LIGHTING);
                    }
                    /* Draw standard dusk time model? */
                    if((obj_ptr->flags & SAR_OBJ_FLAG_HIDE_DUSK_MODEL) ?
                        (scene->tod_code == SARSceneTODCodeDusk) : 1
                    )
                    {
                        /* Disable GL_LIGHTING when drawing dusk time model. */
                        StateGLDisable(&display->state_gl, GL_LIGHTING);
                        SARVisualModelCallList(obj_ptr->visual_model_dusk);
                        StateGLEnable(&display->state_gl, GL_LIGHTING);
                    }
                    /* Draw standard night time model? */
                    if((obj_ptr->flags & SAR_OBJ_FLAG_HIDE_NIGHT_MODEL) ?
                        (scene->tod_code == SARSceneTODCodeNight) : 1
                    )   
                    {
                        /* Disable GL_LIGHTING when drawing night time model. */
                        StateGLDisable(&display->state_gl, GL_LIGHTING);
                        SARVisualModelCallList(obj_ptr->visual_model_night);
                        StateGLEnable(&display->state_gl, GL_LIGHTING);
                    }

		    POST_CALLLIST_RESET_STATES

                    /* Draw lights. */
                    SARDrawLights(
			display, scene,
			obj_ptr->light, obj_ptr->total_lights
		    );
                }
                glPopMatrix();
                break;

              /* ***************************************************** */
              case SAR_OBJ_TYPE_RUNWAY:
                obj_runway_ptr = obj_ptr->data;
                if(obj_runway_ptr == NULL)
                    break;

                /* Skip if camera is under helipad. */
                if(camera_pos.z < pos->z)
                    break;

                /* Get position and check if in range with camera. */
                distance = hypot(pos->x - camera_pos.x, pos->y - camera_pos.y);
                if(distance > obj_ptr->range)
                    break;

                /* Calculate 3D distance. */
                distance3d = hypot(distance, pos->z - camera_pos.z);
        
                /* Get direction. */
                dir = &obj_ptr->dir;
        
                glPushMatrix();
                {
                    /* Translate and rotate. */
                    glTranslatef((GLfloat)pos->x, (GLfloat)pos->z, (GLfloat)-pos->y);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->heading), 0.0, 1.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->pitch), 1.0, 0.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->bank), 0.0, 0.0, 1.0);

                    SARDrawRunway(
			display, scene,
			obj_ptr, obj_runway_ptr, distance3d
		    );
                }
                glPopMatrix();
		POST_CALLLIST_RESET_STATES
                break;

              /* ***************************************************** */
              case SAR_OBJ_TYPE_HELIPAD:
                obj_helipad_ptr = obj_ptr->data;
                if(obj_helipad_ptr == NULL)
                    break;

		/* Skip if camera is under helipad. */
		if(camera_pos.z < pos->z)
		    break;

                /* Get position and check if in range with camera. */
                distance = hypot(pos->x - camera_pos.x, pos->y - camera_pos.y);
                if(distance > obj_ptr->range)
                    break;

                /* Calculate 3D distance. */
                distance3d = hypot(distance, pos->z - camera_pos.z);

                /* Get direction. */
                dir = &obj_ptr->dir;

                glPushMatrix();
                {
                    /* Translate and rotate. */
                    glTranslatef((GLfloat)pos->x, (GLfloat)pos->z, (GLfloat)-pos->y);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->heading), 0.0, 1.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->pitch), 1.0, 0.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->bank), 0.0, 0.0, 1.0);

                    SARDrawHelipad(
			display, scene,
			obj_ptr, obj_helipad_ptr, distance3d
		    );
                }
                glPopMatrix();
		POST_CALLLIST_RESET_STATES
                break;

	      /* ************************************************** */
	      case SAR_OBJ_TYPE_HUMAN:
		obj_human_ptr = obj_ptr->data;
		if(obj_human_ptr == NULL)
		    break;

                /* Get position and check if in range with camera. */
                distance = hypot(pos->x - camera_pos.x, pos->y - camera_pos.y);
                if(distance > obj_ptr->range)
                    break;

                /* Get direction. */
                dir = &obj_ptr->dir;

                glPushMatrix();
                {
                    /* Translate and rotate heading. */
                    glTranslatef((GLfloat)pos->x, (GLfloat)pos->z, (GLfloat)-pos->y);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->heading), 0.0, 1.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->pitch), 1.0, 0.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->bank), 0.0, 0.0, 1.0);

		    SARDrawHuman(display, scene, obj_ptr, obj_human_ptr);
		}
		glPopMatrix();
		POST_CALLLIST_RESET_STATES
		break;

              /* ************************************************** */
              case SAR_OBJ_TYPE_SMOKE:
                /* Get position and check if in range with camera. */
                distance = hypot(pos->x - camera_pos.x, pos->y - camera_pos.y);
                if(distance > obj_ptr->range)
                    break;

                /* Get direction. */
                dir = &obj_ptr->dir;

                glPushMatrix();
                {
		    sar_direction_struct smoke_dir;

                    /* Translate. */
                    glTranslatef((GLfloat)pos->x, (GLfloat)pos->z, (GLfloat)-pos->y);

		    /* Adjust heading to camera heading. */
		    SARDrawGetDirFromPos(
			&camera_pos, pos, &smoke_dir
		    );
		    glRotatef(
			(GLfloat)-SFMRadiansToDegrees(
			    smoke_dir.heading
			), 0.0, 1.0, 0.0
		    );
		    glRotatef(
			(GLfloat)-SFMRadiansToDegrees(
			    smoke_dir.pitch
			), 1.0, 0.0, 0.0
		    );

		    if(obj_ptr->data == NULL)
			break;

		    SARDrawSmoke(
			scene, obj_ptr,
			(sar_object_smoke_struct *)obj_ptr->data
		    );
		}
		glPopMatrix();
		POST_CALLLIST_RESET_STATES
                break;

	      /* ************************************************** */
	      case SAR_OBJ_TYPE_FIRE:

		break;

              /* ************************************************** */
              case SAR_OBJ_TYPE_EXPLOSION:
                /* Get position and check if in range with camera. */
                distance = hypot(pos->x - camera_pos.x, pos->y - camera_pos.y);
                if(distance > obj_ptr->range)
                    break;

                /* Get direction. */
                dir = &obj_ptr->dir;

                glPushMatrix();
                {
                    sar_direction_struct explosion_dir;

                    /* Translate. */
                    glTranslatef((GLfloat)pos->x, (GLfloat)pos->z, (GLfloat)-pos->y);

                    /* Adjust heading to camera heading. */
                    SARDrawGetDirFromPos(
                        &camera_pos, pos, &explosion_dir
                    );
                    glRotatef(
                        (GLfloat)-SFMRadiansToDegrees(
                            explosion_dir.heading
                        ), 0.0, 1.0, 0.0
                    );
                    glRotatef(
                        (GLfloat)-SFMRadiansToDegrees(
                            explosion_dir.pitch
                        ), 1.0, 0.0, 0.0   
                    );

		    if(obj_ptr->data == NULL)
			break;

		    SARDrawExplosion(scene, obj_ptr, obj_ptr->data);
		}
		glPopMatrix();
		POST_CALLLIST_RESET_STATES
		break;

              /* ************************************************** */
              case SAR_OBJ_TYPE_PREMODELED:
                /* Get position and check if in range with camera. */
                distance = hypot(pos->x - camera_pos.x, pos->y - camera_pos.y);
                if(distance > obj_ptr->range)
                    break;

                /* Get direction. */
                dir = &obj_ptr->dir;

                glPushMatrix();
                {
                    /* Translate and rotate. */
                    glTranslatef((GLfloat)pos->x, (GLfloat)pos->z, (GLfloat)-pos->y);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->heading), 0.0, 1.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->pitch), 1.0, 0.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->bank), 0.0, 0.0, 1.0);

		    SARDrawPremodeled(
			display, scene, obj_ptr, obj_ptr->data,
			&camera_pos, &camera_dir
		    );
                }
                glPopMatrix();
		POST_CALLLIST_RESET_STATES
                break;

	      /* **************************************************** */
	      case SAR_OBJ_TYPE_WATERCRAFT:
	      case SAR_OBJ_TYPE_AUTOMOBILE:
	      case SAR_OBJ_TYPE_STATIC:
                /* Get position and check if in range with camera. */
                distance = hypot(pos->x - camera_pos.x, pos->y - camera_pos.y);
                if(distance > obj_ptr->range)
                    break;

                /* Get direction. */
                dir = &obj_ptr->dir;

                glPushMatrix();
                {
                    /* Translate and rotate. */
                    glTranslatef((GLfloat)pos->x, (GLfloat)pos->z, (GLfloat)-pos->y);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->heading), 0.0, 1.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->pitch), 1.0, 0.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->bank), 0.0, 0.0, 1.0);

		    /* Draw standard day time model? */
                    if((obj_ptr->flags & SAR_OBJ_FLAG_HIDE_DAY_MODEL) ?
                        (scene->tod_code == SARSceneTODCodeDay) : 1
                    )
                    {
                        SARVisualModelCallList(obj_ptr->visual_model);
                    }
                    /* Draw standard dawn time model? */
                    if((obj_ptr->flags & SAR_OBJ_FLAG_HIDE_DAWN_MODEL) ?
                        (scene->tod_code == SARSceneTODCodeDawn) : 1
                    )
                    {
                        /* Disable GL_LIGHTING when drawing dawn time model. */
                        StateGLDisable(&display->state_gl, GL_LIGHTING);
                        SARVisualModelCallList(obj_ptr->visual_model_dawn);
                        StateGLEnable(&display->state_gl, GL_LIGHTING);
                    }
                    /* Draw standard dusk time model? */
                    if((obj_ptr->flags & SAR_OBJ_FLAG_HIDE_DUSK_MODEL) ?
                        (scene->tod_code == SARSceneTODCodeDusk) : 1
                    )
                    {
                        /* Disable GL_LIGHTING when drawing dusk time model. */
                        StateGLDisable(&display->state_gl, GL_LIGHTING);
                        SARVisualModelCallList(obj_ptr->visual_model_dusk);
                        StateGLEnable(&display->state_gl, GL_LIGHTING);
                    }
                    /* Draw standard night time model? */
                    if((obj_ptr->flags & SAR_OBJ_FLAG_HIDE_NIGHT_MODEL) ?
                        (scene->tod_code == SARSceneTODCodeNight) : 1
                    )   
                    {
                        /* Disable GL_LIGHTING when drawing night time model. */
                        StateGLDisable(&display->state_gl, GL_LIGHTING);
                        SARVisualModelCallList(obj_ptr->visual_model_night);
                        StateGLEnable(&display->state_gl, GL_LIGHTING);
                    }

		    POST_CALLLIST_RESET_STATES

		    /* Draw lights. */
		    SARDrawLights(
			display, scene,
			obj_ptr->light, obj_ptr->total_lights
		    );
                }
                glPopMatrix();
		break;
            }
        }

	if(option.atmosphere)
	{
            StateGLDisable(&display->state_gl, GL_FOG);
	}

	/* Go to 2D. */
	GWOrtho2D(display);

	StateGLDisable(&display->state_gl, GL_COLOR_MATERIAL);
	StateGLDisable(&display->state_gl, GL_LIGHTING);
	StateGLDisable(&display->state_gl, GL_LIGHT0);
	SAR_DRAW_TEXTURE_1D_OFF
	SAR_DRAW_TEXTURE_2D_OFF
	StateGLDisable(&display->state_gl, GL_ALPHA_TEST);

	/* Draw in cockpit HUD. */
	if(scene->camera_ref == SAR_CAMERA_REF_COCKPIT)
	{
	    SARDrawHUD(
		display,
		scene,
		ptr, total,
		view_aspect
	    );
	}
	else
	{
            SARDrawSetColor(&option.message_color);
	    GWSetFont(display, option.message_font);
	    SARDrawOutsideAttitude(
                display,
                scene,
                scene->player_obj_ptr,
                display->width, display->height
	    );
	}
	/* Draw messages. */
	SARDrawMessages(display, scene);

        /* Draw text input prompt. */
        SARTextInputDraw(core_ptr);

        /* Draw wheel brakes state. */
        switch(player_wheel_brakes)
        {
          case 2:
	    GWSetFont(display, option.message_font);
	    glColor4f(1.0, 0.0, 0.0, 1.0);
	    GWDrawString(
                display,
                10,
		display->height - 20,
                "Parking Brakes On"
            );
	    break;

	  case 1:
            GWSetFont(display, option.message_font);
            glColor4f(
                (GLfloat)MAX((gc == NULL) ? 1.0 : gc->wheel_brakes_coeff, 0.0),
                0.0, 0.0, 1.0
            );
            GWDrawString(
                display,
                10,
                display->height - 20,
                "Wheel Brakes On"
            );
            break;
        }

	/* Draw help. */
	if(core_ptr->display_help)
	    SARDrawHelp(core_ptr);


        /* Put GL buffer to X Window. */
        GWSwapBuffer(display);

        return;
}


/*
 *      Redraws scene as a map.
 */
void SARDrawMap(sar_core_struct *core_ptr, Boolean draw_for_gcc)
{
        int i, n, *total;
        GLfloat view_aspect;
        gw_display_struct *display;
        sar_scene_struct *scene;
        sar_object_struct ***ptr, *obj_ptr;
	sar_color_struct *color_ptr;
        sar_direction_struct *dir;
        sar_position_struct *pos;
	sar_object_aircraft_struct *obj_aircraft_ptr;
	sar_object_helipad_struct *obj_helipad_ptr;
	sar_object_runway_struct *obj_runway_ptr;
        double distance, distance3d, icon_len;


        /* Get initial pointers to resources (valid). */
        display = core_ptr->display;
        scene = core_ptr->scene;
        ptr = &core_ptr->object;
        total = &core_ptr->total_objects;

        /* Reset translations. */
        SARReshapeCB(display, -1, -1, -1, -1);
        if(display->height > 0)
            view_aspect = (GLfloat)display->width / (GLfloat)display->height;
        else
            view_aspect = (800 / 600);
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
	/* Set perspective with zfar to be just a bit farther than
	 * the camera altitude from altitude of 0.
	 */
        gluPerspective(
            SFMRadiansToDegrees(scene->camera_fovz),   /* Field of view. */
            view_aspect,
            1.0,
	    scene->camera_pos.z + 100.0
/*	    SFMFeetToMeters(SAR_MAX_MAP_VIEW_HEIGHT) */
        );
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();


        /* Set camera location, this will set camera position variables
         * for this source file.
         */
        SARSetCamera(scene, ptr, total);

        /* Reset other variables. */
        player_wheel_brakes = 0;


	/* Calculate icon length based on camera z height. */
	icon_len = (scene->camera_fovz * camera_pos.z) * 0.05;

	/* Get pointer to ground base color, going to use this to clear
	 * the frame buffer.
	 */
	color_ptr = &scene->base.color;

        /* Clear GL depth buffer and color, match color to be
	 * the color of the ground base.
	 */
        glClearDepth((GLclampd)1.0);
        glClearColor(
	    (GLfloat)color_ptr->r, (GLfloat)color_ptr->g,
	    (GLfloat)color_ptr->b, (GLfloat)color_ptr->a
	);
        glClear(
            GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT
        );


        /* Turn off depth testing and turn on depth writing for
         * drawing the foundations (the ground base visual models).
         */
        StateGLDisable(&display->state_gl, GL_LIGHTING);
	StateGLDisable(&display->state_gl, GL_LIGHT0);
	StateGLDisable(&display->state_gl, GL_LIGHT1);
	StateGLDisable(&display->state_gl, GL_COLOR_MATERIAL);

	SAR_DRAW_DEPTH_TEST_OFF

	SAR_DRAW_TEXTURE_1D_OFF

	SAR_DRAW_TEXTURE_2D_OFF


	/* Draw ground base. */
        SARDrawSceneFoundations(display, scene, view_aspect);

	/* Textued objects? */
        if(option.textured_objects)
        {
            StateGLEnable(&display->state_gl, GL_ALPHA_TEST);
	    StateGLAlphaFunc(&display->state_gl, GL_GREATER, 0.5);

	    SAR_DRAW_TEXTURE_2D_ON
        }
        else
        {
	    SAR_DRAW_TEXTURE_2D_OFF
        }
        /* Don't use 1D texture. */
	SAR_DRAW_TEXTURE_1D_OFF
            
        /* Begin drawing objects. */
        for(i = 0; i < core_ptr->total_objects; i++)
        {
            obj_ptr = core_ptr->object[i];
            if(obj_ptr == NULL)
                continue;

            /* Check if object is to have depth testing or not. */
/* Disable depth testing for map drawing
            if(obj_ptr->flags & SAR_OBJ_FLAG_NO_DEPTH_TEST)
            {
		SAR_DRAW_DEPTH_TEST_OFF
            }
            else if(!(obj_ptr->flags & SAR_OBJ_FLAG_NO_DEPTH_TEST))
            {
		SAR_DRAW_DEPTH_TEST_ON
            }
 */

	    switch(*(int *)obj_ptr)
	    {
              /* ***************************************************** */
	      case SAR_OBJ_TYPE_STATIC:
              case SAR_OBJ_TYPE_GROUND:
                /* Get position and check if in range with camera. */
                pos = &obj_ptr->pos;
                distance = hypot(pos->x - camera_pos.x, pos->y - camera_pos.y);
                if(distance > obj_ptr->range)
                    break;

                /* Get direction. */
                dir = &obj_ptr->dir;

                glPushMatrix();
                {
                    /* Translate and rotate. */
                    glTranslatef((GLfloat)pos->x, 0, (GLfloat)-pos->y);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->heading), 0.0, 1.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->pitch), 1.0, 0.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->bank), 0.0, 0.0, 1.0);

                    /* Draw standard model if defined. */
                    SARVisualModelCallList(obj_ptr->visual_model);
                }
                glPopMatrix();
		POST_CALLLIST_RESET_STATES
		break;

              /* ***************************************************** */
              case SAR_OBJ_TYPE_AIRCRAFT:
		obj_aircraft_ptr = obj_ptr->data;
		if(obj_aircraft_ptr == NULL)
		    break;

                /* Always silence engine sounds in map view. */
                SARSoundAdjustEngineVolume(
                    core_ptr->recorder,
                    obj_aircraft_ptr->engine_inside_sndplay,
                    obj_aircraft_ptr->engine_outside_sndplay,
                    0, SAR_ENGINE_STATE_OFF,
		    0, 0
                );

                /* Get position and direction, we will be drawing the
		 * intercept lines first unconditionally to range
		 * since we want intercept lines always visable.
		 */
                pos = &obj_ptr->pos;
		dir = &obj_ptr->dir;

		/* Is this the player object? */
		if((obj_ptr == scene->player_obj_ptr) &&
                   (obj_aircraft_ptr->total_intercepts > 0)
		)
		{
		    sar_obj_intercept_struct *intercept_ptr;

		    /* This is the player object, so draw its intercept
		     * points.
		     */
		    glBegin(GL_LINES);
		    {
                        glNormal3f(0, 1, 0);

                        glColor4f((GLfloat)1.0, (GLfloat)0.7, (GLfloat)0.0, (GLfloat)1.0);
			glVertex3f((GLfloat)pos->x, 0, (GLfloat)-pos->y);

			for(n = 0; n < obj_aircraft_ptr->total_intercepts; n++)
			{
			    intercept_ptr = obj_aircraft_ptr->intercept[n];
			    if(intercept_ptr == NULL)
				break;

			    if(n < obj_aircraft_ptr->cur_intercept)
				continue;

                            glVertex3f(
				(GLfloat)intercept_ptr->x, 0, (GLfloat)-intercept_ptr->y
			    );
                            glColor4f(1.0, 0.0, 0.0, 1.0);

			    if(n < (obj_aircraft_ptr->total_intercepts - 1))
                                glVertex3f(
                                    (GLfloat)intercept_ptr->x, 0, (GLfloat)-intercept_ptr->y
                                );
			}
		    }
		    glEnd();
		}

		/* Now perform range check. */
                distance = hypot(pos->x - camera_pos.x, pos->y - camera_pos.y);
                if(distance > obj_ptr->range)
                    break;

		/* Begin drawing aircraft `icon'. */
                glPushMatrix();
                {  
                    /* Translate and rotate. */
                    glTranslatef((GLfloat)pos->x, 0, (GLfloat)-pos->y);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->heading),
			0.0, 1.0, 0.0
		    );

                    V3DTextureSelect(NULL);
                    glShadeModel(GL_FLAT);
                    glBegin(GL_LINES);
                    {
                        glNormal3f(0, 1, 0);

                        if(obj_ptr == scene->player_obj_ptr)
                            glColor4f(1.0, 1.0, 0.0, 1.0);
                        else
                            glColor4f(1.0, 0.0, 0.0, 1.0);

                        glVertex3f(0, 0, 0);
                        glVertex3f(0, 0, (GLfloat)(-(1.0 * icon_len)));

                        glVertex3f((GLfloat)(-0.5 * icon_len), 0, 0);
                        glVertex3f((GLfloat)(0.5 * icon_len), 0, 0);
                    }
                    glEnd();
                }    
                glPopMatrix();
		POST_CALLLIST_RESET_STATES
		break;

              /* ***************************************************** */
              case SAR_OBJ_TYPE_HELIPAD:
                obj_helipad_ptr = obj_ptr->data;
                if(obj_helipad_ptr == NULL)
                    break;

                /* Get position and check if in range with camera. */
                pos = &obj_ptr->pos;
                distance = hypot(pos->x - camera_pos.x, pos->y - camera_pos.y);
                if(distance > obj_ptr->range)
                    break;

                /* Get direction. */
                dir = &obj_ptr->dir;

                glPushMatrix();
                {
		    double x_min, x_max, y_min, y_max;

		    x_min = -(icon_len / 2);
		    x_max = icon_len / 2;
		    y_min = -(icon_len / 2);
                    y_max = icon_len / 2;

                    /* Translate and rotate. */
                    glTranslatef((GLfloat)pos->x, 0, (GLfloat)-pos->y);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->heading), 0.0, 1.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->pitch), 1.0, 0.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->bank), 0.0, 0.0, 1.0);

		    V3DTextureSelect(NULL);
		    glShadeModel(GL_FLAT);
                    glBegin(GL_QUADS);
                    {
                        glNormal3f(0, 1, 0);

                        glColor4f(0.0, 0.0, 1.0, 1.0);
                        
                        glVertex3f((GLfloat)x_min, 0, (GLfloat)-y_max);
                        glVertex3f((GLfloat)x_min, 0, (GLfloat)-y_min);
                        glVertex3f((GLfloat)x_max, 0, (GLfloat)-y_min);
                        glVertex3f((GLfloat)x_max, 0, (GLfloat)-y_max);

                        glColor4f(1.0, 1.0, 1.0, 1.0);

                        glVertex3f((GLfloat)(x_min * 0.8), 0, (GLfloat)(-y_max * 0.2));
                        glVertex3f((GLfloat)(x_min * 0.8), 0, (GLfloat)(-y_min * 0.2));
                        glVertex3f((GLfloat)(x_max * 0.8), 0, (GLfloat)(-y_min * 0.2));
                        glVertex3f((GLfloat)(x_max * 0.8), 0, (GLfloat)(-y_max * 0.2));

                        glVertex3f((GLfloat)(x_min * 0.2), 0, (GLfloat)(-y_max * 0.8));
                        glVertex3f((GLfloat)(x_min * 0.2), 0, (GLfloat)(-y_min * 0.8));
                        glVertex3f((GLfloat)(x_max * 0.2), 0, (GLfloat)(-y_min * 0.8));
                        glVertex3f((GLfloat)(x_max * 0.2), 0, (GLfloat)(-y_max * 0.8));
		    }
		    glEnd();

                }
                glPopMatrix();
		POST_CALLLIST_RESET_STATES
                break;

              /* ***************************************************** */
              case SAR_OBJ_TYPE_RUNWAY:
                obj_runway_ptr = obj_ptr->data;
                if(obj_runway_ptr == NULL)
                    break;
                    
                /* Get position and check if in range with camera. */     
                pos = &obj_ptr->pos;
                distance = hypot(pos->x - camera_pos.x, pos->y - camera_pos.y);
                if(distance > obj_ptr->range)
                    break;

                /* Calculate 3D distance. */
                distance3d = hypot(distance, pos->z - camera_pos.z);

                /* Get direction. */
                dir = &obj_ptr->dir;

                glPushMatrix();
                {
                    double x_min, x_max, y_min, y_max;

		    x_max = MAX(obj_runway_ptr->width / 2, icon_len / 2);
                    x_min = -x_max;
		    y_max = MAX(obj_runway_ptr->length / 2, icon_len / 2);
                    y_min = -y_max;

                    /* Translate and rotate. */
                    glTranslatef((GLfloat)pos->x, 0, (GLfloat)-pos->y);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->heading), 0.0, 1.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->pitch), 1.0, 0.0, 0.0);
                    glRotatef((GLfloat)-SFMRadiansToDegrees(dir->bank), 0.0, 0.0, 1.0);

                    V3DTextureSelect(NULL);
                    glShadeModel(GL_FLAT);
                    glBegin(GL_QUADS);
                    {
                        glNormal3f(0, 1, 0);

                        glColor4f(0.0, 0.0, 1.0, 1.0);
            
                        glVertex3f((GLfloat)x_min, 0, (GLfloat)-y_max);
                        glVertex3f((GLfloat)x_min, 0, (GLfloat)-y_min);
                        glVertex3f((GLfloat)x_max, 0, (GLfloat)-y_min);   
                        glVertex3f((GLfloat)x_max, 0, (GLfloat)-y_max);

                        glColor4f(1.0, 1.0, 1.0, 1.0);

                        glVertex3f((GLfloat)(x_min * 0.5), 0, (GLfloat)(-(y_max + (x_max * 0.5))));
                        glVertex3f((GLfloat)(x_min * 0.5), 0, (GLfloat)(-(y_min - (x_max * 0.5))));
                        glVertex3f((GLfloat)(x_max * 0.5), 0, (GLfloat)(-(y_min - (x_max * 0.5))));
                        glVertex3f((GLfloat)(x_max * 0.5), 0, (GLfloat)(-(y_max + (x_max * 0.5))));
                    }
                    glEnd();
                }
                glPopMatrix();
		POST_CALLLIST_RESET_STATES
                break;
	    }
        }

        /* Go to 2D. */
        GWOrtho2D(display);

        StateGLDisable(&display->state_gl, GL_ALPHA_TEST);
	StateGLDisable(&display->state_gl, GL_COLOR_MATERIAL);
	SAR_DRAW_TEXTURE_1D_OFF
	SAR_DRAW_TEXTURE_2D_OFF


	/* Draw attitude and stats of player object. */
        SARDrawSetColor(&option.message_color);
        GWSetFont(display, option.message_font);
        SARDrawOutsideAttitude(
            display,
            scene,
            scene->player_obj_ptr,
            display->width, display->height
        );

        /* Draw messages. */
        SARDrawMessages(display, scene);

	/* Draw text input prompt. */
	SARTextInputDraw(core_ptr);

        /* Draw help. */
        if(core_ptr->display_help)   
            SARDrawHelp(core_ptr);   


        /* Put GL buffer to X Window. */
        GWSwapBuffer(display);
          
        return;
}           
