#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <math.h>

#include "sfm.h"

#include "obj.h"
#include "messages.h"
#include "explosion.h"
#include "simmanage.h"
#include "simcb.h"
#include "simop.h"
#include "simutils.h"
#include "sarsound.h"
#include "sar.h"


void SARSimInitModelCB(
        void *realm_ptr, SFMModelStruct *model,
        void *client_data
);
void SARSimDestroyModelCB(
        void *realm_ptr, SFMModelStruct *model,
        void *client_data                     
);

void SARSimAirborneCB(
        void *realm_ptr, SFMModelStruct *model,
        void *client_data
);
void SARSimTouchDownCB(
	void *realm_ptr, SFMModelStruct *model,
	void *client_data, double impact_coeff
);
void SARSimCollisionCB(
        void *realm_ptr, SFMModelStruct *model, SFMModelStruct *obstruction,
        void *client_data, double impact_coeff 
);
void SARSimObjectCollisionCB(
        void *client_data,              /* Core structure. */
        sar_object_struct *obj_ptr, sar_object_struct *obstruction_obj_ptr,
        double impact_coeff
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))


/*
 *	FDM initialize callback, sets up initial values from
 *	corresponding object (ie position and landed state values).
 */
void SARSimInitModelCB(
        void *realm_ptr, SFMModelStruct *model,
        void *client_data
)
{
	int obj_num;
	sar_object_struct *obj_ptr;
	SFMRealmStruct *realm = (SFMRealmStruct *)realm_ptr;
	sar_core_struct *core_ptr = (sar_core_struct *)client_data;
	if((realm == NULL) ||
           (model == NULL) ||
           (core_ptr == NULL)
	)
	    return;

        /* Match object from FDM. */
        obj_ptr = SARSimMatchObjectFromFDM(
            core_ptr->object, core_ptr->total_objects,
            model, &obj_num
        );
        if(obj_ptr == NULL)
            return;

	/* Set up initial values, note that position and direction
	 * values will probably be reset soon afterwards since
	 * creation of FDM's usually happen in the middle of
	 * scene file parsings and their positions will be
	 * set after creation of the FDM.
	 */
/*
	model->position.x = obj_ptr->pos.x;
        model->position.y = obj_ptr->pos.y;
        model->position.z = obj_ptr->pos.z;

        model->direction.heading = obj_ptr->dir.heading;
        model->direction.pitch = obj_ptr->dir.pitch;
        model->direction.bank = obj_ptr->dir.bank;
 */

	/* Always assume landed. */
	model->landed_state = True;

	return;
}


/*
 *	FDM destroy callback.
 */
void SARSimDestroyModelCB(
        void *realm_ptr, SFMModelStruct *model,
        void *client_data
)
{
	int i;
	sar_object_struct *obj_ptr;
	sar_object_aircraft_struct *obj_aircraft_ptr;
        SFMRealmStruct *realm = (SFMRealmStruct *)realm_ptr;
        sar_core_struct *core_ptr = (sar_core_struct *)client_data;
        if((realm == NULL) ||    
           (model == NULL) ||
           (core_ptr == NULL)   
        )   
            return;

	/* Note that model input is considered invalid, do not
	 * referance! So scan through objects on core structure and
	 * unset fdm.
	 */
	for(i = 0; i < core_ptr->total_objects; i++)
	{
	    obj_ptr = core_ptr->object[i];
	    if(obj_ptr == NULL)
		continue;

	    switch(obj_ptr->type)
	    {
	      case SAR_OBJ_TYPE_AIRCRAFT:
                obj_aircraft_ptr = (sar_object_aircraft_struct *)obj_ptr->data;
                if(obj_aircraft_ptr == NULL)
                    break;

		if(obj_aircraft_ptr->fdm == model)
		    obj_aircraft_ptr->fdm = NULL;

		break;
	    }
	}


	return;
}


/*
 *	FDM airborne callback, whenever the FDM just leaves the ground.
 */
void SARSimAirborneCB(
        void *realm_ptr, SFMModelStruct *model,
        void *client_data
)
{
        SFMRealmStruct *realm = (SFMRealmStruct *)realm_ptr;
        sar_core_struct *core_ptr = (sar_core_struct *)client_data;
        if((realm == NULL) ||
           (model == NULL) ||
           (core_ptr == NULL)
        )
            return;




        return;
}       


/*
 *	FDM touchdown callback, whenever the FDM just lands.
 */
void SARSimTouchDownCB(
        void *realm_ptr, SFMModelStruct *model,
        void *client_data, double impact_coeff
)
{
        int n, obj_num, crash_cause = 0;
        double distance_to_camera;		/* In meters. */
        sar_scene_struct *scene;
        sar_position_struct *pos, *cam_pos;
	sar_position_struct *vel = NULL;
	sar_direction_struct *dir;
	sar_object_struct *obj_ptr;
        sar_object_aircraft_struct *obj_aircraft_ptr = NULL;
	sar_contact_bounds_struct *cb;
	sar_obj_flags_t crash_flags = 0;
        double contact_radius = 0.0;
	int air_worthy_state = SAR_AIR_WORTHY_FLYABLE;

        SFMRealmStruct *realm = (SFMRealmStruct *)realm_ptr;
        sar_core_struct *core_ptr = (sar_core_struct *)client_data;
        if((realm == NULL) ||
           (model == NULL) ||  
           (core_ptr == NULL)
        )
            return;

	scene = core_ptr->scene;
	if(scene == NULL)
	    return;


#define DO_PLAY_LAND_BELLY              \
{ \
 SARSoundPlayPredefined( \
  core_ptr->recorder, \
  SAR_SOUND_PREDEFINED_LAND_BELLY, \
  scene, distance_to_camera \
 ); \
}

#define DO_PLAY_LAND_SKI                \
{ \
 SARSoundPlayPredefined( \
  core_ptr->recorder, \
  SAR_SOUND_PREDEFINED_LAND_SKI, \
  scene, distance_to_camera \
 ); \
}

#define DO_PLAY_LAND_SKI_SKID           \
{ \
 SARSoundPlayPredefined( \
  core_ptr->recorder, \
  SAR_SOUND_PREDEFINED_LAND_SKI_SKID, \
  scene, distance_to_camera \
 ); \
}

#define DO_PLAY_LAND_WHEEL_SKID         \
{ \
 SARSoundPlayPredefined( \
  core_ptr->recorder, \
  SAR_SOUND_PREDEFINED_LAND_WHEEL_SKID, \
  scene, distance_to_camera \
 ); \
}

#define DO_PLAY_CRASH_GROUND            \
{ \
 SARSoundPlayPredefined( \
  core_ptr->recorder, \
  SAR_SOUND_PREDEFINED_CRASH_GROUND, \
  scene, distance_to_camera \
 ); \
}

	/* Match object from FDM. */
	obj_ptr = SARSimMatchObjectFromFDM(
	    core_ptr->object, core_ptr->total_objects,
	    model, &obj_num
	);
	if(obj_ptr == NULL)
	    return;


        /* Get positions and velocities. */
	pos = &obj_ptr->pos;
        pos->x = model->position.x;
	pos->y = model->position.y;
	pos->z = model->position.z;

	cam_pos = &scene->camera_pos;

	dir = &obj_ptr->dir;
	dir->heading = model->direction.heading;
        dir->pitch = model->direction.pitch;
        dir->bank = model->direction.bank;

	cb = obj_ptr->contact_bounds;
	if(cb != NULL)
	{
	    crash_flags = cb->crash_flags;
	    contact_radius = SARSimGetFlatContactRadius(obj_ptr);
	}


        /* Calculate distance between object and camera. */
        distance_to_camera = hypot(hypot(
            pos->x - cam_pos->x,
            pos->y - cam_pos->y
            ), pos->z - cam_pos->z
        );

        /* Get object type specific values. */
        switch(obj_ptr->type)
        {
          case SAR_OBJ_TYPE_AIRCRAFT:
            obj_aircraft_ptr = (sar_object_aircraft_struct *)
		obj_ptr->data;
            if(obj_aircraft_ptr == NULL)
                break;

            vel = &obj_aircraft_ptr->vel;
            vel->x = model->velocity_vector.x;
            vel->y = model->velocity_vector.y;
            vel->z = model->velocity_vector.z;
 
            air_worthy_state = obj_aircraft_ptr->air_worthy_state;
            break;

	}

  
        /* If object was already in flight but was not flyable
         * (ie it hit a building and is falling to the ground),
         * then contacting the ground is a crash.
         */
        if(air_worthy_state != SAR_AIR_WORTHY_FLYABLE)
        {
            crash_cause = 3;  /* Crashed already, just contacted ground. */
        }
	/* Impact tolorance exceeded? */ 
        else if(impact_coeff > 1.0)
        {
            crash_cause = 1;  /* Impact tolorance exceeded. */
        }
        /* Contacted ground at bad angle? */
        else
        {
            if(dir->pitch > (1.0 * PI))
            {
                if(dir->pitch < (1.75 * PI))
                    crash_cause = 2;
            }
            else
            {
                if(dir->pitch > (0.25 * PI))
                    crash_cause = 2;
            }
            
            if(dir->bank > (1.0 * PI))
            {
                if(dir->bank < (1.75 * PI))
                    crash_cause = 2;
            }
            else
            {
                if(dir->bank > (0.25 * PI))
                    crash_cause = 2;
            }
        }
/*
printf("Touch down %i %f\n", crash_cause, impact_coeff);
 */
        /* ******************************************************** */
        /* Was there a crash? */
        if(crash_cause)
        {
            double explosion_radius = 10.0;
            sar_position_struct explosion_pos;
            char text[1024];

            /* Set object `as crashed' by its type. */
            switch(obj_ptr->type)
            {
              case SAR_OBJ_TYPE_AIRCRAFT:
		SARSimSetAircraftCrashed(
		    scene, &core_ptr->object, &core_ptr->total_objects,
		    obj_ptr, (sar_object_aircraft_struct *)obj_ptr->data
		);
                break;

/* Add support for other types of objects that can crash. */

            }

            /* Is this the player object? */
            if(obj_ptr == scene->player_obj_ptr)
            {
                /* Print crash message. */
                switch(crash_cause)
                {
                  case 2:
                    strcpy(text, SAR_MESG_CRASH_ROTATION_TOO_STEEP);
                    SARMessageAdd(scene, text);
                    break;

                  case 1:
		    /* There should be exactly one occurance of "%.0f%%"
		     * in SAR_MESG_CRASH_IMPACTED_PAST_TOLORANCE for the
		     * substitution to work.
		     */
                    sprintf(
			text,
			SAR_MESG_CRASH_IMPACTED_PAST_TOLORANCE,
                        (double)(impact_coeff * 100)
                    );
                    SARMessageAdd(scene, text);
                    break;
                }
            }
            else
            {
                /* An object other than the player object has
                 * crashed.
                 */
		char numstr[256];

		sprintf(numstr, "Object #%i", obj_num);

		(*text) = '\0';
		strcat(text, "*** ");
		strncat(text, ((obj_ptr->name == NULL) ?
		    numstr : obj_ptr->name),
		    80
		);
		strcat(text, " has crashed! ***");

                SARMessageAdd(scene, text);
            }

            /* Begin creating explosion. */

            /* Set position and size for creation of explosion. */
            memcpy(&explosion_pos, &obj_ptr->pos, sizeof(sar_position_struct));
            explosion_radius = MAX(contact_radius * 1.4, 10.0);

            /* Create explosion and go to external view (as needed).
             * If the object is not set crashable, it implies its already
             * crashed into something and so we should not create another
             * explosion or change view.
             */
            if(crash_flags & SAR_CRASH_FLAG_CRASH_OTHER)
            {
                /* Just crashed into ground, was not previously hit. */
                sar_object_struct *obj_ptr2;

		/* Create explosion object. */
                n = ExplosionCreatePosition(
                    scene,
                    &core_ptr->object, &core_ptr->total_objects,
                    &explosion_pos,     /* Position of explosion. */
                    explosion_radius,   /* Radius of size in meters. */
                    obj_num,            /* Referance object number. */
                    SAR_STD_TEXNAME_EXPLOSION
                );
                if(n > -1)
                {
                    obj_ptr2 = core_ptr->object[n];
                    if(obj_ptr2 != NULL)
                    {
                        obj_ptr2->life_span = cur_millitime +
                            option.crash_explosion_life_span;
                    }
		}

                /* Is this the player object? */
                if(scene->player_obj_ptr == obj_ptr)
                {
                    /* Set spot position. */
                    scene->camera_ref = SAR_CAMERA_REF_SPOT;
                    scene->camera_target = scene->player_obj_num;
                }
            }

            /* Play explosion sound. */
            DO_PLAY_CRASH_GROUND

            /* Call mission destroy notify to let it know this
             * object has crashed.
             */
            SARMissionDestroyNotify(core_ptr, obj_ptr);
        }
        else
        {
            /* No, not crashed. */
            sar_obj_landing_gear_struct *lgear_ptr = NULL;


            /* Handle safe landing by object type. */
            switch(obj_ptr->type)
            {
              case SAR_OBJ_TYPE_AIRCRAFT:
                obj_aircraft_ptr = (sar_object_aircraft_struct *)
		    obj_ptr->data;
                if(obj_aircraft_ptr == NULL)
                    break;

                if(obj_aircraft_ptr->total_landing_gears > 0)
                    lgear_ptr = obj_aircraft_ptr->landing_gear[0];

                /* Landed on belly? */
                if((lgear_ptr == NULL) ? 1 :
                    (lgear_ptr->flags & SAR_LGEAR_FLAG_STATE)
                )
                {
                    /* Landed on belly. */
                    if(obj_aircraft_ptr->speed > SFMMPHToMPC(5))
                    {
                        DO_PLAY_LAND_BELLY
                    }
                    else
                    {
                        DO_PLAY_LAND_SKI
                    }
                }
                else
                {
                    /* Landed on gear. */

                    /* Landing gear a ski? */
                    if((lgear_ptr == NULL) ? 1 :
                        (lgear_ptr->flags & SAR_LGEAR_FLAG_SKI)
                    )
                    {
                        /* Landing gear is a ski. */
                        if(obj_aircraft_ptr->speed > SFMMPHToMPC(5))
                        {
                            DO_PLAY_LAND_SKI_SKID
                        } 
                        else
                        {
                            DO_PLAY_LAND_SKI
                        }
                    }
                    else
                    {
                        /* Landing gear is a wheel. */

                        /* Brakes on? */  
                        if(obj_aircraft_ptr->wheel_brakes_state != 0)
                        {
                            if(obj_aircraft_ptr->speed > SFMMPHToMPC(10))
                            {
                                DO_PLAY_LAND_WHEEL_SKID
                            }
                        }
                        else
                        {
                            if(obj_aircraft_ptr->speed > SFMMPHToMPC(25))
                            {
                                DO_PLAY_LAND_WHEEL_SKID
                            }
                        }
                    }
                }       /* Landed on gear. */
                break;


            }   /* Handle safe landing by object type. */


	    /* Call mission land notify. */
	    SARMissionLandNotify(core_ptr, obj_ptr);

        }


#undef DO_PLAY_LAND_BELLY
#undef DO_PLAY_LAND_SKI
#undef DO_PLAY_LAND_SKI_SKID
#undef DO_PLAY_LAND_WHEEL_SKID
#undef DO_PLAY_CRASH_GROUND

	return;
}


/*
 *	FDM midair collision callback. This function won't be called
 *	by the SFM library (since all crash detection values passed
 *	to the SFM are False), instead SAR will do collision detection
 *	and call this function.
 */
void SARSimCollisionCB(
        void *realm_ptr, SFMModelStruct *model, SFMModelStruct *obstruction,
        void *client_data, double impact_coeff
)
{

/* Ignore for now. */

	return;
}

/*
 *      Same as SARSimCollisionCB except that it takes inputs in
 *	the form of object pointers.
 */
void SARSimObjectCollisionCB(
	void *client_data,		/* Core structure. */
	sar_object_struct *obj_ptr, sar_object_struct *obstruction_obj_ptr,
        double impact_coeff
)
{
        int n, ref_object = -1;
        double r, distance_to_camera;
        double  tower_offset_x = 50.0,
                tower_offset_y = 50.0,
                tower_offset_z = 50.0;
        sar_position_struct *pos, *camera_pos, explosion_pos;
        sar_scene_struct *scene;
        sar_object_struct *obj_ptr2;
        sar_object_aircraft_struct *obj_aircraft_ptr;
	sar_contact_bounds_struct *cb;
        sar_core_struct *core_ptr = (sar_core_struct *)client_data; 
        if((core_ptr == NULL) ||
           (obj_ptr == NULL) ||
           (obstruction_obj_ptr == NULL)
        )
            return;

        scene = core_ptr->scene;
        if(scene == NULL)
            return;

#define DO_PLAY_CRASH_OBSTRUCTION	\
{ \
 SARSoundPlayPredefined( \
  core_ptr->recorder, \
  SAR_SOUND_PREDEFINED_CRASH_OBSTRUCTION, \
  scene, distance_to_camera \
 ); \
}

	/* Get pointer to position of victim object. */
        pos = &obj_ptr->pos;

	/* Get pointer to position of camera. */
        camera_pos = &scene->camera_pos;

        /* Calculate distance to camera. */
        distance_to_camera = hypot(hypot(
            pos->x - camera_pos->x,
            pos->y - camera_pos->y
            ), pos->z - camera_pos->z
        );


	/* Check if obstruction object (the `crashed into' object) is
	 * valid.
	 */
        if(obstruction_obj_ptr != NULL)
        {
            double bearing;     	/* Victim obj to obstruction obj. */
            double theta, speed, d;
            sar_position_struct *pos2, *vel;


	    /* Get pointer to obstruction object's position. */
            pos2 = &obstruction_obj_ptr->pos;

            /* Begin setting up victim object's values to mark that it has
             * crashed into the obstruction object.
             */

	    /* Change some values on victim object by its type, do not
	     * set object as fully crashed since it's probably plumitting
	     * to the ground.
	     */
            switch(obj_ptr->type)
            {
              case SAR_OBJ_TYPE_AIRCRAFT:
               obj_aircraft_ptr = (sar_object_aircraft_struct *)obj_ptr->data;
                if(obj_aircraft_ptr == NULL)
                    break;

                vel = &obj_aircraft_ptr->vel;
                speed = obj_aircraft_ptr->speed;

                /* Calculate bearing to obstruction. */
                bearing = SFMSanitizeRadians(
                    (PI / 2) - atan2(pos2->y - pos->y, pos2->x - pos->x)
                );

                /* Calculate impact of velocity loss. */
                theta = bearing - obj_ptr->dir.heading;
                vel->y = vel->y - (speed * cos(theta));
                vel->x = 0;

		/* Get size of object. */
		d = SARSimGetFlatContactRadius(obj_ptr) * 4.0;
		if(d < 10.0)
		    d = 10.0;
                tower_offset_y = -cos(bearing) * d;
                tower_offset_z = d;

                /* Set air worthy state based on if object is already
		 * landed or not.
		 */
                if(obj_aircraft_ptr->landed)
                {
		    /* Call SARSimSetAircraftCrashed() procedure since
		     * aircraft is already on ground and won't crash
		     * again.
		     */
		    SARSimSetAircraftCrashed(
			scene, &core_ptr->object, &core_ptr->total_objects,
			obj_ptr, obj_aircraft_ptr
		    );
                }
                else
                {
		    /* Aircraft is still flying, mark it as out of
		     * control.
		     */
                    obj_aircraft_ptr->air_worthy_state =
                        SAR_AIR_WORTHY_OUT_OF_CONTROL;

		    /* Turn engines off. */
		    obj_aircraft_ptr->engine_state = SAR_ENGINE_STATE_OFF;
                }
                break;

/* Add support for other objects that can crash into other objects. */

            }
        }	/* if(obstruction_obj_ptr != NULL) */


        /* Get pointer to victim object's contact bounds structure. */
	cb = obj_ptr->contact_bounds;
	if(cb != NULL)
	{
	    /* Remove SAR_CRASH_FLAG_CRASH_OTHER flag from victim
	     * object's crash flags so that the victim object won't
	     * crash into other objects from now on.
	     */
	    cb->crash_flags &= ~SAR_CRASH_FLAG_CRASH_OTHER;
	}


	/* Begin creating explosion. */

        /* Get position and values for creating explosion object. */
        memcpy(&explosion_pos, &obj_ptr->pos, sizeof(sar_position_struct));
        ref_object = SARGetObjectNumberFromPointer(
            scene, core_ptr->object, core_ptr->total_objects, obj_ptr
        );
	r = SARSimGetFlatContactRadius(obj_ptr) * 1.4;
        if(r < 10.0)
            r = 10.0;

	/* Create explosion object. */
        n = ExplosionCreatePosition(
            scene, &core_ptr->object, &core_ptr->total_objects,
            &explosion_pos,     /* Position of explosion. */
            r,                  /* Radius of size in meters. */
            ref_object,         /* Referance object number. */
            SAR_STD_TEXNAME_EXPLOSION
        );
        if(n > -1)  
        {
            obj_ptr2 = core_ptr->object[n];
            if(obj_ptr2 != NULL)
            {
               obj_ptr2->life_span = cur_millitime +
                    option.crash_explosion_life_span;
            }
        }

	/* Play crash obstruction sound. */
        DO_PLAY_CRASH_OBSTRUCTION


        /* Is this the player object? */
        if(scene->player_obj_ptr == obj_ptr)
        {         
            char text[128];


            /* Set tower position. */
            scene->camera_pos.x = obj_ptr->pos.x +
                tower_offset_x;
            scene->camera_pos.y = obj_ptr->pos.y +
                tower_offset_y;
            scene->camera_pos.z = obj_ptr->pos.z +
                tower_offset_z;

            scene->camera_ref = SAR_CAMERA_REF_TOWER;
            scene->camera_target = scene->player_obj_num;


            /* Report crash type. */ 
            if(obstruction_obj_ptr != NULL)
            {
		cb = obstruction_obj_ptr->contact_bounds;
                switch((cb == NULL) ? 0 : cb->crash_type)
                {
                  case SAR_CRASH_TYPE_OBSTRUCTION:
                    strcpy(text, SAR_MESG_CRASH_OBSTRUCTION);
                    break;

                  case SAR_CRASH_TYPE_GROUND:
                    /* Never should occure for midair collisions. */
                    strcpy(text, SAR_MESG_CRASH_GROUND);
                    break;

                  case SAR_CRASH_TYPE_MOUNTAIN:
                    strcpy(text, SAR_MESG_CRASH_MOUNTAIN);
                    break;

                  case SAR_GRASH_TYPE_BUILDING:
                    strcpy(text, SAR_MESG_CRASH_BOULDING);
                    break;

                  case SAR_GRASH_TYPE_AIRCRAFT:
                    strcpy(text, SAR_MESG_CRASH_AIRCRAFT);
                    break;
                }
                SARMessageAdd(scene, text);
            }
        }

#undef DO_PLAY_CRASH_OBSTRUCTION

	return;
}


